import datetime
import logging
from collections.abc import Mapping
from typing import Any

from fme.core.coordinates import (
    HorizontalCoordinates,
    NullVerticalCoordinate,
    SerializableHorizontalCoordinates,
    SerializableVerticalCoordinate,
    VerticalCoordinate,
)
from fme.core.dataset.data_typing import VariableMetadata
from fme.core.dataset.utils import decode_timestep, encode_timestep
from fme.core.gridded_ops import GriddedOperations
from fme.core.mask_provider import MaskProvider, MaskProviderABC, NullMaskProvider


class MissingDatasetInfo(ValueError):
    def __init__(self, info: str):
        super().__init__(
            f"Dataset used for initialization is missing required information: {info}"
        )


class IncompatibleDatasetInfo(ValueError):
    pass


class DatasetInfo:
    """
    Information about a dataset.

    Generally this is used to provide static information about a training dataset
    when initializing a model, and to validate the compatibility of a model's training
    dataset against inference datasets.

    Its API is meant to support type-safe attribute access in cases where data
    may or may not be available. If data is not available, a MissingDatasetInfo
    exception is raised.
    """

    def __init__(
        self,
        horizontal_coordinates: HorizontalCoordinates | None = None,
        vertical_coordinate: VerticalCoordinate | None = None,
        mask_provider: MaskProvider | None = None,
        timestep: datetime.timedelta | None = None,
        variable_metadata: Mapping[str, VariableMetadata] | None = None,
        gridded_operations: GriddedOperations | None = None,
        img_shape: tuple[int, int] | None = None,
    ):
        self._horizontal_coordinates = horizontal_coordinates
        self._vertical_coordinate = vertical_coordinate
        self._mask_provider = mask_provider
        self._timestep = timestep
        self._variable_metadata = variable_metadata

        # The gridded_operations and img_shape arguments are only provided for backwards
        # compatibility with older serialized states that have these attributes instead
        # of horizontal_coordinates.
        if gridded_operations is not None and horizontal_coordinates is not None:
            msg = "Cannot provide both gridded_operations and horizontal_coordinates. "
            raise ValueError(msg)
        self._gridded_operations = gridded_operations
        if img_shape is not None and horizontal_coordinates is not None:
            msg = "Cannot provide both img_shape and horizontal_coordinates. "
            raise ValueError(msg)
        self._img_shape = img_shape

    def __eq__(self, other) -> bool:
        if not isinstance(other, DatasetInfo):
            return False
        return (
            self._img_shape == other._img_shape
            and self._gridded_operations == other._gridded_operations
            and self._horizontal_coordinates == other._horizontal_coordinates
            and self._vertical_coordinate == other._vertical_coordinate
            and self._mask_provider == other._mask_provider
            and self._timestep == other._timestep
            and self._variable_metadata == other._variable_metadata
        )

    def __repr__(self) -> str:
        return (
            f"DatasetInfo( "
            f"horizontal_coordinates={self._horizontal_coordinates}, "
            f"vertical_coordinate={self._vertical_coordinate}, "
            f"timestep={self._timestep}), "
            f"mask_provider={self._mask_provider}, "
            f"variable_metadata={self._variable_metadata})"
        )

    def assert_compatible_with(self, other: "DatasetInfo"):
        issues = []
        if (
            self._horizontal_coordinates is not None
            and other._horizontal_coordinates is not None
        ):
            if self._horizontal_coordinates != other._horizontal_coordinates:
                issues.append(
                    f"horizontal_coordinates is not compatible, "
                    f"{self._horizontal_coordinates} != {other._horizontal_coordinates}"
                )
        if (
            self._vertical_coordinate is not None
            and other._vertical_coordinate is not None
        ) and not (
            isinstance(self._vertical_coordinate, NullVerticalCoordinate)
            or isinstance(other._vertical_coordinate, NullVerticalCoordinate)
        ):
            if self._vertical_coordinate != other._vertical_coordinate:
                issues.append(
                    f"vertical_coordinate is not compatible, "
                    f"{self._vertical_coordinate} != {other._vertical_coordinate}"
                )
        if self._mask_provider is not None and other._mask_provider is not None:
            if self._mask_provider != other._mask_provider:
                issues.append(
                    f"mask_provider is not compatible, "
                    f"{self._mask_provider} != {other._mask_provider}"
                )
        if self._timestep is not None:
            if self._timestep != other._timestep:
                issues.append(
                    f"timestep is not compatible, {self._timestep} != {other._timestep}"
                )
        metadata_conflicts = get_keys_with_conflicts(
            self._variable_metadata, other._variable_metadata
        )
        for key, (a, b) in metadata_conflicts.items():
            logging.warning(
                "DatasetInfo has different metadata from other DatasetInfo for key "
                f"{key}: {a} != {b}"
            )
        if issues:
            raise IncompatibleDatasetInfo(
                "DatasetInfo is not compatible with other DatasetInfo:\n"
                + "\n".join(issues)
            )

    @property
    def img_shape(self) -> tuple[int, int]:
        if self._img_shape is not None:
            return self._img_shape
        if self._horizontal_coordinates is None:
            raise MissingDatasetInfo("horizontal_coordinates")
        result = self._horizontal_coordinates.shape[-2:]
        assert len(result) == 2  # for a happy mypy
        return result

    @property
    def gridded_operations(self) -> GriddedOperations:
        if self._gridded_operations is not None:
            return self._gridded_operations
        if self._horizontal_coordinates is None:
            raise MissingDatasetInfo("horizontal_coordinates")
        if self._mask_provider is None:
            mp: MaskProviderABC = NullMaskProvider
        else:
            mp = self._mask_provider
        return self._horizontal_coordinates.get_gridded_operations(mask_provider=mp)

    @property
    def horizontal_coordinates(self) -> HorizontalCoordinates:
        if self._horizontal_coordinates is None:
            raise MissingDatasetInfo("horizontal_coordinates")
        return self._horizontal_coordinates

    @property
    def vertical_coordinate(self) -> VerticalCoordinate:
        if self._vertical_coordinate is None:
            raise MissingDatasetInfo("vertical_coordinate")
        return self._vertical_coordinate

    @property
    def mask_provider(self) -> MaskProvider:
        if self._mask_provider is None:
            raise MissingDatasetInfo("mask_provider")
        return self._mask_provider

    @property
    def timestep(self) -> datetime.timedelta:
        if self._timestep is None:
            raise MissingDatasetInfo("timestep")
        return self._timestep

    @property
    def variable_metadata(self) -> Mapping[str, VariableMetadata]:
        if self._variable_metadata is None:
            return {}
        return self._variable_metadata

    def update_variable_metadata(
        self, new_metadata: Mapping[str, VariableMetadata] | None
    ) -> "DatasetInfo":
        """
        Return a new DatasetInfo with the variable metadata updated.
        """
        return DatasetInfo(
            horizontal_coordinates=self._horizontal_coordinates,
            vertical_coordinate=self._vertical_coordinate,
            mask_provider=self._mask_provider,
            timestep=self._timestep,
            variable_metadata=new_metadata,
            gridded_operations=self._gridded_operations,
            img_shape=self._img_shape,
        )

    def to_state(self) -> dict[str, Any]:
        if self._gridded_operations is not None:
            gridded_operations = self._gridded_operations.to_state()
        else:
            gridded_operations = None
        if self._img_shape is not None:
            img_shape = self._img_shape
        else:
            img_shape = None
        if self._horizontal_coordinates is None:
            horizontal_coordinates = None
        else:
            horizontal_coordinates = self._horizontal_coordinates.to_state()
        if self._vertical_coordinate is None:
            vertical_coordinate = None
        else:
            vertical_coordinate = self._vertical_coordinate.as_dict()
        if self._mask_provider is None:
            mask_provider = None
        else:
            mask_provider = self._mask_provider.to_state()
        if self._timestep is None:
            timestep = None
        else:
            timestep = encode_timestep(self._timestep)
        return {
            "horizontal_coordinates": horizontal_coordinates,
            "vertical_coordinate": vertical_coordinate,
            "mask_provider": mask_provider,
            "timestep": timestep,
            "variable_metadata": self._variable_metadata,
            "gridded_operations": gridded_operations,
            "img_shape": img_shape,
        }

    @classmethod
    def from_state(cls, state: dict[str, Any]) -> "DatasetInfo":
        if state.get("gridded_operations") is not None:
            # this is for backwards compatibility with older serialized states
            assert state.get("horizontal_coordinates") is None
            gridded_ops = GriddedOperations.from_state(state["gridded_operations"])
        else:
            gridded_ops = None
        if state.get("img_shape") is not None:
            # this is for backwards compatibility with older serialized states
            assert state.get("horizontal_coordinates") is None
            img_shape = state["img_shape"]
        else:
            img_shape = None
        if state.get("horizontal_coordinates") is not None:
            horizontal_coordinates = SerializableHorizontalCoordinates.from_state(
                state["horizontal_coordinates"]
            )
        else:
            horizontal_coordinates = None
        if state.get("vertical_coordinate") is not None:
            vertical_coordinate = SerializableVerticalCoordinate.from_state(
                state["vertical_coordinate"]
            )
        else:
            vertical_coordinate = None
        if state.get("mask_provider") is not None:
            mask_provider = MaskProvider.from_state(state["mask_provider"])
        else:
            mask_provider = None
        if state.get("timestep") is not None:
            timestep = decode_timestep(state["timestep"])
        else:
            timestep = None
        variable_metadata = state.get("variable_metadata")
        return cls(
            horizontal_coordinates=horizontal_coordinates,
            vertical_coordinate=vertical_coordinate,
            mask_provider=mask_provider,
            timestep=timestep,
            variable_metadata=variable_metadata,
            gridded_operations=gridded_ops,
            img_shape=img_shape,
        )


def get_keys_with_conflicts(
    a: Mapping[str, VariableMetadata] | None,
    b: Mapping[str, VariableMetadata] | None,
) -> Mapping[str, tuple[VariableMetadata, VariableMetadata]]:
    """
    Get the return the keys where the values in the two dictionaries are different.
    """
    a = a or {}
    b = b or {}
    keys_with_conflicts = {}
    for key in a.keys() & b.keys():
        if a[key] != b[key]:
            keys_with_conflicts[key] = (a[key], b[key])
    return keys_with_conflicts



import asyncio
import datetime
from collections.abc import Hashable, MutableMapping, Sequence

import numpy as np
import torch
import xarray as xr
import zarr

from fme.core.coordinates import (
    HEALPixCoordinates,
    HorizontalCoordinates,
    LatLonCoordinates,
)
from fme.core.dataset.config import FillNaNsConfig

SLICE_NONE = slice(None)


def _get_indexers(
    variable: xr.Variable, dims: Sequence[Hashable]
) -> tuple[slice | None, ...]:
    """Returns a tuple of indexers for the dimensions provided.

    Indexers select all data from dimensions that exist in the variable, and
    create new axes for dimensions that do not exist. The returned tuple will
    have the same length as the provided sequence of dimensions.

    Inspired by similar code in xarray:
    https://github.com/pydata/xarray/blob/1d43672574332615f225089d69f95a9f8d81d912/xarray/core/computation.py#L681-L688
    """
    indexers: list[slice | None] = []
    for dim in dims:
        if dim in variable.dims:
            indexers.append(SLICE_NONE)
        else:
            indexers.append(None)
    return tuple(indexers)


def as_broadcasted_tensor(
    variable: xr.Variable,
    dims: Sequence[Hashable],
    shape: Sequence[int],
) -> torch.tensor:
    """Load data from variable and broadcast to tensor with the given shape."""
    arr = variable.values
    indexers = _get_indexers(variable, dims)
    tensor = torch.as_tensor(arr[indexers])
    return torch.broadcast_to(tensor, shape)


def _broadcast_array_to_tensor(
    array: np.ndarray,
    dims: Sequence[Hashable],
    shape: Sequence[int],
) -> torch.tensor:
    """Convert from numpy array to tensor and broadcast to the given shape.

    Note: if number of dimensions is different between input array and
    desired shape, it is assumed that array only has time dimension and will be
    broadcasted over all spatial dimensions.
    """
    tensor = torch.as_tensor(array)
    if len(array.shape) != len(shape):
        if array.shape[0] != shape[0]:
            raise ValueError("Must have matching time dimension")
        if len(array.shape) != 1:
            raise ValueError("If broadcasting, array must be 1D")
        # insert new singleton dimensions for missing spatial dimensions
        n_spatial_dims = len(dims) - 1
        tensor = tensor[(...,) + (None,) * n_spatial_dims]
    return torch.broadcast_to(tensor, shape)


async def _get_item(group, name, selection):
    async_array = await group.getitem(name)
    if len(async_array.shape) != len(selection):
        if len(async_array.shape) != 1:
            raise ValueError(
                f"Index selection slices/indices were provided as {selection} "
                f"but array {name} has {len(async_array.shape)} dimensions. "
            )
        else:
            # 1d scalar arrays are sliced along time dim
            array_selection = selection[0]
    else:
        array_selection = selection
    return await async_array.getitem(array_selection)


async def _get_items(url, names, selection):
    async_group = await zarr.api.asynchronous.open(store=url)
    coroutines = []
    for name in names:
        coroutines.append(_get_item(async_group, name, selection))
    return await asyncio.gather(*coroutines)


def _load_all_variables_zarr_async(
    path: str, variables: Sequence[str], selection: tuple[slice | int, ...]
) -> MutableMapping[str, np.ndarray]:
    """Load data from a variables into memory.

    Assumes that path contains a zarr group and uses async.

    Assumes that the time dimension is the first dimension in the dataset.
    """
    loop = asyncio.get_event_loop()
    arrays = loop.run_until_complete(_get_items(path, variables, selection))
    return {k: v for k, v in zip(variables, arrays)}


def _load_all_variables(
    ds: xr.Dataset, variables: Sequence[str], time_slice: slice = SLICE_NONE
) -> xr.Dataset:
    """Load data from a variables into memory.

    This function leverages xarray's lazy loading to load only the time slice
    (or chunk[s] for the time slice) of the variables we need.
    """
    if "time" in ds.dims:
        ds = ds.isel(time=time_slice)
    return ds[variables].compute()


def load_series_data_zarr_async(
    idx: int,
    n_steps: int,
    path: str,
    names: list[str],
    final_dims: list[str],
    final_shape: list[int],
    nontime_selection: tuple[int | slice, ...],
    fill_nans: FillNaNsConfig | None = None,
):
    time_slice = slice(idx, idx + n_steps)
    selection = (time_slice, *nontime_selection)
    loaded = _load_all_variables_zarr_async(path, names, selection)
    if fill_nans is not None:
        for k, v in loaded.items():
            loaded[k] = np.nan_to_num(v, nan=fill_nans.value)
    arrays = {}
    for name in names:
        arrays[name] = _broadcast_array_to_tensor(loaded[name], final_dims, final_shape)
    return arrays


def load_series_data(
    idx: int,
    n_steps: int,
    ds: xr.Dataset,
    names: list[str],
    final_dims: list[str],
    final_shape: list[int],
    fill_nans: FillNaNsConfig | None = None,
):
    time_slice = slice(idx, idx + n_steps)
    loaded = _load_all_variables(ds, names, time_slice)
    # Fill NaNs after subsetting time slice to avoid triggering loading all
    # data, since we do not use dask.
    if fill_nans is not None:
        loaded = loaded.fillna(fill_nans.value)
    arrays = {}
    for n in names:
        variable = loaded[n].variable
        if len(variable.shape) != len(final_shape):
            if variable.shape[0] != final_shape[0]:
                raise ValueError("Must have matching time dimension")
            if len(variable.shape) != 1:
                raise ValueError("If broadcasting, array must be 1D")
        arrays[n] = as_broadcasted_tensor(variable, final_dims, final_shape)
    return arrays


def get_horizontal_coordinates(
    ds: xr.Dataset,
    spatial_dimensions: str,
    dtype: torch.dtype | None,
) -> tuple[HorizontalCoordinates, list[str]]:
    """Return the horizontal coordinate class and dimension names."""
    min_ndim = 3 if spatial_dimensions == "latlon" else 4
    coords: HorizontalCoordinates
    for da in ds.data_vars.values():
        if da.ndim >= min_ndim:
            dims = list(da.dims)
            break
    if spatial_dimensions == "latlon":
        lat_name, lon_name = dims[-2:]
        coords = LatLonCoordinates(
            lon=torch.tensor(ds[lon_name].values, dtype=dtype),
            lat=torch.tensor(ds[lat_name].values, dtype=dtype),
        )
        return coords, dims[-2:]
    elif spatial_dimensions == "healpix":
        face_name, height_name, width_name = dims[-3:]
        coords = HEALPixCoordinates(
            face=torch.tensor(ds[face_name].values, dtype=dtype),
            height=torch.tensor(ds[height_name].values, dtype=dtype),
            width=torch.tensor(ds[width_name].values, dtype=dtype),
        )
        return coords, dims[-3:]
    else:
        raise ValueError(
            f"spatial_dimensions must be either 'latlon' or 'healpix', "
            f"but got {spatial_dimensions}"
        )


def decode_timestep(microseconds: int) -> datetime.timedelta:
    return datetime.timedelta(microseconds=microseconds)


def encode_timestep(timedelta: datetime.timedelta) -> int:
    return timedelta // datetime.timedelta(microseconds=1)


def get_nonspacetime_dimensions(
    ds: xr.Dataset, horizontal_dims: list[str]
) -> list[str]:
    """Get all dimensions that are not time or horizontal dimensions."""
    nonspacetime_dims: list[str] = []
    dim_order: list[str] = []

    for da in ds.data_vars.values():
        if da.ndim >= len(horizontal_dims) + 1 and da.ndim > len(dim_order):
            dim_order = list(da.dims)
    for dim in dim_order:
        if dim not in horizontal_dims and dim != "time":
            nonspacetime_dims.append(dim)
    return nonspacetime_dims



from collections.abc import Sequence
from datetime import timedelta

import cftime
import pytest
import torch
import xarray as xr

from .time_coarsen import coarsen_batch

DIM_SIZES = (2, 4, 2, 2)  # n_samples_in_batch, n_timesteps_in_window, n_lat, n_lon
VARNAME = "foo"


def get_windowed_batch(dim_sizes: Sequence[int], start_time: Sequence[int]):
    n_samples, n_timesteps, n_lat, n_lon = dim_sizes
    # get data varying only along time dim
    data = (
        torch.arange(n_timesteps, dtype=torch.float64)
        .repeat(n_samples, n_lat, n_lon, 1)
        .movedim(3, 1)
    )
    target = {VARNAME: data}
    time = get_batch_time(n_timesteps, start_time, n_samples=n_samples)
    return target, time


def get_batch_time(
    n_timesteps: int,
    start_time: Sequence[int],
    n_samples: int = 2,
    freq_hrs: int = 6,
):
    sample_times = [
        _get_time_array(start_time, n_timesteps, start_n_offset, freq_hrs)
        for start_n_offset in range(n_samples)
    ]
    return xr.DataArray(
        data=xr.concat(sample_times, dim="sample"),
        dims=["sample", "time"],
    )


def _get_time_array(
    start_time: Sequence[int], n_timesteps: int, start_n_offset: int, freq_hrs: int
):
    # start each sample one timestep behind the last
    start_time = cftime.DatetimeJulian(*start_time) + timedelta(
        hours=6 * start_n_offset
    )
    time_index = xr.date_range(
        start=start_time,
        periods=n_timesteps,
        freq=f"{freq_hrs}h",
        calendar="julian",
        use_cftime=True,
    )
    return xr.DataArray(data=time_index, dims=["time"]).drop_vars(["time"])


@pytest.mark.parametrize(
    [
        "coarsen_factor",
        "start_timestep",
        "expected_coarsened_data",
        "expected_coarsened_time",
        "expected_coarsened_start_timestep",
    ],
    [
        pytest.param(
            1,
            0,
            [0.0, 1.0, 2.0, 3.0],
            get_batch_time(start_time=(2020, 1, 1, 0, 0, 0), n_timesteps=4),
            0,
            id="coarsen_factor_1",
        ),
        pytest.param(
            2,
            0,
            [0.5, 2.5],
            get_batch_time(
                start_time=(2020, 1, 1, 3, 0, 0), n_timesteps=2, freq_hrs=12
            ),
            0,
            id="coarsen_factor_2",
        ),
        pytest.param(
            2,
            3,
            [0.5, 2.5],
            get_batch_time(
                start_time=(2020, 1, 1, 3, 0, 0), n_timesteps=2, freq_hrs=12
            ),
            1,
            id="coarsen_factor_2_start_timestep_3",
        ),
        pytest.param(
            4,
            0,
            [1.5],
            get_batch_time(
                start_time=(2020, 1, 1, 9, 0, 0), n_timesteps=1, freq_hrs=24
            ),
            0,
            id="coarsen_factor_4",
        ),
    ],
)
def test_time_coarsen(
    coarsen_factor: int,
    start_timestep: int,
    expected_coarsened_data: Sequence[float],
    expected_coarsened_time: Sequence[cftime.DatetimeJulian],
    expected_coarsened_start_timestep: int,
    dim_sizes: Sequence[int] = DIM_SIZES,
):
    target, time = get_windowed_batch(
        dim_sizes=dim_sizes, start_time=(2020, 1, 1, 0, 0, 0)
    )
    (
        target_coarsened,
        coarsened_start_timestep,
        time_coarsened,
    ) = coarsen_batch(
        data=target,
        start_timestep=start_timestep,
        batch_time=time,
        coarsen_factor=coarsen_factor,
    )
    # check the coarsened data time dim size
    assert target_coarsened[VARNAME].size(dim=1) == len(
        expected_coarsened_data
    ), "target coarsened time dim"
    assert time_coarsened.sizes["time"] == len(
        expected_coarsened_data
    ), "time coarsened time dim"
    # check the coarsened data values
    n_samples, _, n_lat, n_lon = dim_sizes
    (
        torch.testing.assert_close(
            target_coarsened[VARNAME],
            torch.tensor(expected_coarsened_data, dtype=torch.float64)
            .repeat(n_samples, n_lat, n_lon, 1)
            .movedim(3, 1),
        ),
        "target coarsened value",
    )
    # check the coarsened start timestep
    assert coarsened_start_timestep == expected_coarsened_start_timestep
    # check the coarsened data time coordinate values
    (
        xr.testing.assert_allclose(time_coarsened, expected_coarsened_time),
        "time initial condition value",
    )



import dataclasses

from torch import nn

from fme.ace.registry.registry import ModuleConfig, ModuleSelector


@ModuleSelector.register("prebuilt")
@dataclasses.dataclass
class PreBuiltBuilder(ModuleConfig):
    """
    A simple module configuration which returns a pre-defined module.

    Used mainly for testing.
    """

    module: nn.Module

    def build(
        self,
        n_in_channels: int,
        n_out_channels: int,
        img_shape: tuple[int, int],
    ) -> nn.Module:
        return self.module



import dataclasses
import warnings
from collections.abc import Sequence

from fme.ace.data_loading.augmentation import AugmentationConfig
from fme.core.dataset.config import (
    ConcatDatasetConfig,
    MergeDatasetConfig,
    XarrayDataConfig,
)
from fme.core.distributed import Distributed


@dataclasses.dataclass
class DataLoaderConfig:
    r"""
    Configuration for the data loader.

    Note: Setting `time_buffer` to a value greater than 0 results in pre-loading
        samples of length `time_buffer + n_timesteps_required`, where
        `n_timesteps_required` is the number of timesteps required for training
        the model (initial condition(s) plus forward step(s)). These pre-loaded samples
        become a window from which samples of the required length are drawn without
        replacement. The windows will overlap by an amount such that no samples are
        skipped, with exception of the last window, which is dropped if incomplete.
        This is useful for improving data loading throughput and reducing the number of
        reads. There must be enough pre-loaded samples in the dataset to produce at
        least one batch at the configured batch size. Independent data will be seen
        every `time_buffer + 1` batches, i.e., this is the number of samples in each
        pre-loaded window.

    Parameters:
        dataset: Could be a single dataset configuration,
            or a sequence of datasets to be concatenated using the keyword `concat`,
            or datasets from different sources to be merged using the keyword `merge`.
            For backwards compatibility, it can also be a sequence of
            datasets, which will be concatenated.
            During `merge`, if multiple datasets contain the same data variable,
            the version from the first source is loaded and other sources are ignored.
        batch_size: Number of samples per batch.
        num_data_workers: Number of parallel workers to use for data loading.
        prefetch_factor: how many batches a single data worker will attempt to
            hold in host memory at a given time.
        augmentation: Configuration for data augmentation.
        sample_with_replacement: If provided, the dataset will be
            sampled randomly with replacement to the given size each period,
            instead of retrieving each sample once (either shuffled or not).
        time_buffer: How many more continuous timesteps to load in memory than the
            required number of timesteps for a single batch. Setting this to greater
            than 0 should improve data loading performance, however, it also decreases
            the independence of subsequent batches if shuffled batches are desired.
    """

    dataset: (
        ConcatDatasetConfig
        | MergeDatasetConfig
        | XarrayDataConfig
        | Sequence[XarrayDataConfig]
    )
    batch_size: int
    num_data_workers: int = 0
    prefetch_factor: int | None = None
    augmentation: AugmentationConfig = dataclasses.field(
        default_factory=lambda: AugmentationConfig()
    )
    sample_with_replacement: int | None = None
    time_buffer: int = 0

    def __post_init__(self):
        dist = Distributed.get_instance()
        if self.batch_size % dist.world_size != 0:
            raise ValueError(
                "batch_size must be divisible by the number of parallel "
                f"workers, got {self.batch_size} and {dist.world_size}"
            )
        # TODO: remove following backwards compatibility code in a future release
        if isinstance(self.dataset, Sequence):
            warnings.warn(
                "Dataset list format is deprecated. "
                "Use `concat` to specify concatenating datasets.",
                DeprecationWarning,
            )
            self.dataset = ConcatDatasetConfig(concat=self.dataset)
        self._zarr_engine_used = self.dataset.zarr_engine_used
        if self.time_buffer < 0:
            raise ValueError(
                "time_buffer must be greater than or equal to 0. "
                f"Got {self.time_buffer}"
            )

    @property
    def zarr_engine_used(self) -> bool:
        """
        Whether any of the configured datasets are using the Zarr engine.
        """
        return self._zarr_engine_used



# flake8: noqa
# fmt: off
# isort: skip_file

"""
This file contains a fix that we needed to get the SFNO to work on multiple
unroll steps in multiprocessing (e.g. multi-GPU mode.) We forked this code from
the torch harmonics sht.py file [*].

[*] https://github.com/NVIDIA/torch-harmonics/blob/17eefa53468d1a885d72087918eba905fa53e10a/torch_harmonics/sht.py
"""


# coding=utf-8

# SPDX-FileCopyrightText: Copyright (c) 2022 The torch-harmonics Authors. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import numpy as np
import torch
import torch.nn as nn
import torch.fft

from torch_harmonics.quadrature import legendre_gauss_weights, lobatto_weights, clenshaw_curtiss_weights
from torch_harmonics.legendre import _precompute_legpoly
import torch_harmonics

from fme.core.device import get_device

class RealSHT(nn.Module):
    """
    Defines a module for computing the forward (real-valued) SHT.
    Precomputes Legendre Gauss nodes, weights and associated Legendre polynomials on these nodes.
    The SHT is applied to the last two dimensions of the input

    [1] Schaeffer, N. Efficient spherical harmonic transforms aimed at pseudospectral numerical simulations, G3: Geochemistry, Geophysics, Geosystems.
    [2] Wang, B., Wang, L., Xie, Z.; Accurate calculation of spherical and vector spherical harmonic expansions via spectral element grids; Adv Comput Math.
    """

    def __init__(self, nlat, nlon, lmax=None, mmax=None, grid="lobatto", norm="ortho", csphase=True):
        """
        Initializes the SHT Layer, precomputing the necessary quadrature weights

        Parameters:
        nlat: input grid resolution in the latitudinal direction
        nlon: input grid resolution in the longitudinal direction
        grid: grid in the latitude direction (for now only tensor product grids are supported)
        """

        super().__init__()

        self.nlat = nlat
        self.nlon = nlon
        self.grid = grid
        self.norm = norm
        self.csphase = csphase

        # TODO: include assertions regarding the dimensions

        # compute quadrature points
        if self.grid == "legendre-gauss":
            cost, w = legendre_gauss_weights(nlat, -1, 1)
            self.lmax = lmax or self.nlat
        elif self.grid == "lobatto":
            cost, w = lobatto_weights(nlat, -1, 1)
            self.lmax = lmax or self.nlat-1
        elif self.grid == "equiangular":
            cost, w = clenshaw_curtiss_weights(nlat, -1, 1)
            # cost, w = fejer2_weights(nlat, -1, 1)
            self.lmax = lmax or self.nlat
        elif self.grid == "healpix":
            raise(NotImplementedError("'healpix' grid not supported by InverseRealVectorSHT"))
        else:
            raise(ValueError("Unknown quadrature mode"))

        # apply cosine transform and flip them
        tq = np.flip(np.arccos(cost))

        # determine the dimensions
        self.mmax = mmax or self.nlon // 2 + 1

        # combine quadrature weights with the legendre weights
        weights = torch.from_numpy(w)
        pct = torch.as_tensor(_precompute_legpoly(self.mmax, self.lmax, tq, norm=self.norm, csphase=self.csphase))
        weights = torch.einsum('mlk,k->mlk', pct, weights)

        # remember quadrature weights
        self.weights = weights.float().to(get_device())

    def extra_repr(self):
        """
        Pretty print module
        """
        return f'nlat={self.nlat}, nlon={self.nlon},\n lmax={self.lmax}, mmax={self.mmax},\n grid={self.grid}, csphase={self.csphase}'

    def forward(self, x: torch.Tensor):

        assert(x.shape[-2] == self.nlat)
        assert(x.shape[-1] == self.nlon)

        # apply real fft in the longitudinal direction
        x = 2.0 * torch.pi * torch.fft.rfft(x, dim=-1, norm="forward")

        # do the Legendre-Gauss quadrature
        x = torch.view_as_real(x)

        # distributed contraction: fork
        out_shape = list(x.size())
        out_shape[-3] = self.lmax
        out_shape[-2] = self.mmax
        xout = torch.zeros(out_shape, dtype=x.dtype, device=x.device)

        # contraction
        weights = self.weights.to(x.device).to(x.dtype)
        xout[..., 0] = torch.einsum('...km,mlk->...lm', x[..., :self.mmax, 0], weights)
        xout[..., 1] = torch.einsum('...km,mlk->...lm', x[..., :self.mmax, 1], weights)
        x = torch.view_as_complex(xout)

        return x

class InverseRealSHT(nn.Module):
    """
    Defines a module for computing the inverse (real-valued) SHT.
    Precomputes Legendre Gauss nodes, weights and associated Legendre polynomials on these nodes.
    nlat, nlon: Output dimensions
    lmax, mmax: Input dimensions (spherical coefficients). For convenience, these are inferred from the output dimensions

    [1] Schaeffer, N. Efficient spherical harmonic transforms aimed at pseudospectral numerical simulations, G3: Geochemistry, Geophysics, Geosystems.
    [2] Wang, B., Wang, L., Xie, Z.; Accurate calculation of spherical and vector spherical harmonic expansions via spectral element grids; Adv Comput Math.
    """

    def __init__(self, nlat, nlon, lmax=None, mmax=None, grid="lobatto", norm="ortho", csphase=True):

        super().__init__()

        self.nlat = nlat
        self.nlon = nlon
        self.grid = grid
        self.norm = norm
        self.csphase = csphase

        # compute quadrature points
        if self.grid == "legendre-gauss":
            cost, _ = legendre_gauss_weights(nlat, -1, 1)
            self.lmax = lmax or self.nlat
        elif self.grid == "lobatto":
            cost, _ = lobatto_weights(nlat, -1, 1)
            self.lmax = lmax or self.nlat-1
        elif self.grid == "equiangular":
            cost, _ = clenshaw_curtiss_weights(nlat, -1, 1)
            self.lmax = lmax or self.nlat
        elif self.grid == "healpix":
            raise(NotImplementedError("'healpix' grid not supported by RealVectorSHT"))
        else:
            raise(ValueError("Unknown quadrature mode"))

        # apply cosine transform and flip them
        t = np.flip(np.arccos(cost))

        # determine the dimensions
        self.mmax = mmax or self.nlon // 2 + 1

        pct = torch.as_tensor(_precompute_legpoly(self.mmax, self.lmax, t, norm=self.norm, inverse=True, csphase=self.csphase))

        # register buffer
        self.pct = pct.float().to(get_device())

    def extra_repr(self):
        """
        Pretty print module
        """
        return f'nlat={self.nlat}, nlon={self.nlon},\n lmax={self.lmax}, mmax={self.mmax},\n grid={self.grid}, csphase={self.csphase}'

    def forward(self, x: torch.Tensor):

        assert(x.shape[-2] == self.lmax)
        assert(x.shape[-1] == self.mmax)

        # Evaluate associated Legendre functions on the output nodes
        x = torch.view_as_real(x)

        pct = self.pct.to(x.device).to(x.dtype)
        rl = torch.einsum('...lm, mlk->...km', x[..., 0], pct )
        im = torch.einsum('...lm, mlk->...km', x[..., 1], pct )
        xs = torch.stack((rl, im), -1)

        # apply the inverse (real) FFT
        x = torch.view_as_complex(xs)
        x = torch.fft.irfft(x, n=self.nlon, dim=-1, norm="forward")

        return x

torch_harmonics.RealSHT = RealSHT
torch_harmonics.InverseRealSHT = InverseRealSHT



# flake8: noqa
# Copied from https://github.com/ai2cm/modulus/commit/22df4a9427f5f12ff6ac891083220e7f2f54d229
# Copyright (c) 2023, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
import math
from typing import Any, List, Optional, Tuple

import torch
import torch.nn as nn

# get spectral transforms from torch_harmonics
import torch_harmonics as th
import torch_harmonics.distributed as thd
from torch.utils.checkpoint import checkpoint

from .initialization import trunc_normal_

# wrap fft, to unify interface to spectral transforms
# import global convolution and non-linear spectral layers
# helpers
from .layers import (
    MLP,
    ConditionalLayerNorm,
    Context,
    ContextConfig,
    DropPath,
    RealFFT2,
    SpectralAttention2d,
)
from .s2convolutions import SpectralAttentionS2, SpectralConvS2


# heuristic for finding theta_cutoff
def _compute_cutoff_radius(nlat, kernel_shape, basis_type):
    theta_cutoff_factor = {
        "piecewise linear": 0.5,
        "morlet": 0.5,
        "zernike": math.sqrt(2.0),
    }

    return (
        (kernel_shape[0] + 1)
        * theta_cutoff_factor[basis_type]
        * math.pi
        / float(nlat - 1)
    )


class DiscreteContinuousConvS2(nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.conv = th.DiscreteContinuousConvS2(*args, **kwargs)

    def forward(self, x):
        return self.conv(x), x


class SpectralFilterLayer(nn.Module):
    """Spectral filter layer"""

    def __init__(
        self,
        forward_transform,
        inverse_transform,
        embed_dim,
        filter_type="linear",
        operator_type="block-diagonal",
        sparsity_threshold=0.0,
        use_complex_kernels=True,
        hidden_size_factor=1,
        rank=1.0,
        factorization=None,
        separable=False,
        complex_network=True,
        complex_activation="real",
        spectral_layers=1,
        drop_rate=0.0,
        filter_residual=False,
    ):
        super(SpectralFilterLayer, self).__init__()

        if filter_type == "non-linear" and (
            isinstance(forward_transform, th.RealSHT)
            or isinstance(forward_transform, thd.DistributedRealSHT)
        ):
            self.filter = SpectralAttentionS2(
                forward_transform,
                inverse_transform,
                embed_dim,
                operator_type=operator_type,
                sparsity_threshold=sparsity_threshold,
                hidden_size_factor=hidden_size_factor,
                complex_activation=complex_activation,
                spectral_layers=spectral_layers,
                drop_rate=drop_rate,
                bias=False,
            )

        elif filter_type == "non-linear" and (isinstance(forward_transform, RealFFT2)):
            self.filter = SpectralAttention2d(
                forward_transform,
                inverse_transform,
                embed_dim,
                sparsity_threshold=sparsity_threshold,
                hidden_size_factor=hidden_size_factor,
                complex_activation=complex_activation,
                spectral_layers=spectral_layers,
                drop_rate=drop_rate,
                bias=False,
            )

        # spectral transform is passed to the module
        elif filter_type == "linear" and (
            isinstance(forward_transform, th.RealSHT)
            or isinstance(forward_transform, thd.DistributedRealSHT)
        ):
            self.filter = SpectralConvS2(
                forward_transform,
                inverse_transform,
                embed_dim,
                embed_dim,
                operator_type=operator_type,
                rank=rank,
                factorization=factorization,
                separable=separable,
                bias=True,
                use_tensorly=False if factorization is None else True,
                filter_residual=filter_residual,
            )

        elif filter_type == "local":
            # heuristic for finding theta_cutoff
            theta_cutoff = 2 * _compute_cutoff_radius(
                nlat=forward_transform.nlat,
                kernel_shape=(3, 3),
                basis_type="morlet",
            )
            self.filter = DiscreteContinuousConvS2(
                embed_dim,
                embed_dim,
                in_shape=(forward_transform.nlat, forward_transform.nlon),
                out_shape=(inverse_transform.nlat, inverse_transform.nlon),
                kernel_shape=(3, 3),
                basis_type="morlet",
                basis_norm_mode="mean",
                groups=1,
                grid_in=forward_transform.grid,
                grid_out=inverse_transform.grid,
                bias=False,
                theta_cutoff=theta_cutoff,
            )
        else:
            raise (NotImplementedError)

    def forward(self, x):
        return self.filter(x)


class FourierNeuralOperatorBlock(nn.Module):
    """Fourier Neural Operator Block"""

    def __init__(
        self,
        forward_transform,
        inverse_transform,
        embed_dim,
        context_config: ContextConfig,
        filter_type="linear",
        operator_type="diagonal",
        global_layer_norm: bool = False,
        mlp_ratio=2.0,
        drop_rate=0.0,
        drop_path=0.0,
        act_layer=nn.GELU,
        sparsity_threshold=0.0,
        use_complex_kernels=True,
        rank=1.0,
        factorization=None,
        separable=False,
        inner_skip="linear",
        outer_skip=None,  # None, nn.linear or nn.Identity
        concat_skip=False,
        use_mlp=False,
        complex_network=True,
        complex_activation="real",
        spectral_layers=1,
        checkpointing=0,
        filter_residual=False,
    ):
        super(FourierNeuralOperatorBlock, self).__init__()

        self.input_shape_loc = (forward_transform.nlat, forward_transform.nlon)
        self.output_shape_loc = (inverse_transform.nlat, inverse_transform.nlon)

        # norm layer
        self.norm0 = ConditionalLayerNorm(
            embed_dim,
            img_shape=self.input_shape_loc,
            global_layer_norm=global_layer_norm,
            context_config=context_config,
        )

        # convolution layer
        self.filter = SpectralFilterLayer(
            forward_transform,
            inverse_transform,
            embed_dim,
            filter_type,
            operator_type,
            sparsity_threshold,
            use_complex_kernels=use_complex_kernels,
            hidden_size_factor=mlp_ratio,
            rank=rank,
            factorization=factorization,
            separable=separable,
            complex_network=complex_network,
            complex_activation=complex_activation,
            spectral_layers=spectral_layers,
            drop_rate=drop_rate,
            filter_residual=filter_residual,
        )

        if inner_skip == "linear":
            self.inner_skip = nn.Conv2d(embed_dim, embed_dim, 1, 1)
        elif inner_skip == "identity":
            self.inner_skip = nn.Identity()

        self.concat_skip = concat_skip

        if concat_skip and inner_skip is not None:
            self.inner_skip_conv = nn.Conv2d(2 * embed_dim, embed_dim, 1, bias=False)

        if filter_type == "linear" or filter_type == "real linear":
            self.act_layer = act_layer()

        # dropout
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

        # norm layer
        self.norm1 = ConditionalLayerNorm(
            embed_dim,
            img_shape=self.output_shape_loc,
            global_layer_norm=global_layer_norm,
            context_config=context_config,
        )

        if use_mlp == True:
            MLPH = MLP
            mlp_hidden_dim = int(embed_dim * mlp_ratio)
            self.mlp = MLPH(
                in_features=embed_dim,
                hidden_features=mlp_hidden_dim,
                act_layer=act_layer,
                drop_rate=drop_rate,
                checkpointing=checkpointing,
            )

        if outer_skip == "linear":
            self.outer_skip = nn.Conv2d(embed_dim, embed_dim, 1, 1)
        elif outer_skip == "identity":
            self.outer_skip = nn.Identity()

        if concat_skip and outer_skip is not None:
            self.outer_skip_conv = nn.Conv2d(2 * embed_dim, embed_dim, 1, bias=False)

    def forward(self, x, context_embedding):
        x_norm = torch.zeros_like(x)
        x_norm[..., : self.input_shape_loc[0], : self.input_shape_loc[1]] = self.norm0(
            x[..., : self.input_shape_loc[0], : self.input_shape_loc[1]],
            context_embedding,
        )
        x, residual = self.filter(x_norm)

        if hasattr(self, "inner_skip"):
            if self.concat_skip:
                x = torch.cat((x, self.inner_skip(residual)), dim=1)
                x = self.inner_skip_conv(x)
            else:
                x = x + self.inner_skip(residual)

        if hasattr(self, "act_layer"):
            x = self.act_layer(x)

        x_norm = torch.zeros_like(x)
        x_norm[..., : self.output_shape_loc[0], : self.output_shape_loc[1]] = (
            self.norm1(
                x[..., : self.output_shape_loc[0], : self.output_shape_loc[1]],
                context_embedding,
            )
        )
        x = x_norm

        if hasattr(self, "mlp"):
            x = self.mlp(x)

        x = self.drop_path(x)

        if hasattr(self, "outer_skip"):
            if self.concat_skip:
                x = torch.cat((x, self.outer_skip(residual)), dim=1)
                x = self.outer_skip_conv(x)
            else:
                x = x + self.outer_skip(residual)

        return x


class SphericalFourierNeuralOperatorNet(torch.nn.Module):
    """
    Spherical Fourier Neural Operator Network

    Parameters
    ----------
    params : dict
        Dictionary of parameters
    spectral_transform : str, optional
        Type of spectral transformation to use, by default "sht"
    filter_type : str, optional
        Type of filter to use ('linear', 'non-linear'), by default "non-linear"
    operator_type : str, optional
        Type of operator to use ('diaginal', 'dhconv'), by default "diagonal"
    img_shape : tuple, optional
        Shape of the input channels, by default (721, 1440)
    scale_factor : int, optional
        Scale factor to use, by default 16
    in_chans : int, optional
        Number of input channels, by default 2
    out_chans : int, optional
        Number of output channels, by default 2
    embed_dim : int, optional
        Dimension of the embeddings, by default 256
    context_config : ContextConfig, optional
        Context configuration, by default
        ContextConfig(embed_dim_scalar=0, embed_dim_2d=0)
    num_layers : int, optional
        Number of layers in the network, by default 12
    use_mlp : int, optional
        Whether to use MLP, by default True
    mlp_ratio : int, optional
        Ratio of MLP to use, by default 2.0
    activation_function : str, optional
        Activation function to use, by default "gelu"
    encoder_layers : int, optional
        Number of layers in the encoder, by default 1
    pos_embed : bool, optional
        Whether to use positional embedding, by default True
    drop_rate : float, optional
        Dropout rate, by default 0.0
    drop_path_rate : float, optional
        Dropout path rate, by default 0.0
    num_blocks : int, optional
        Number of blocks in the network, by default 16
    sparsity_threshold : float, optional
        Threshold for sparsity, by default 0.0
    hard_thresholding_fraction : float, optional
        Fraction of hard thresholding to apply, by default 1.0
    use_complex_kernels : bool, optional
        Whether to use complex kernels, by default True
    big_skip : bool, optional
        Whether to use big skip connections, by default True
    rank : float, optional
        Rank of the approximation, by default 1.0
    factorization : Any, optional
        Type of factorization to use, by default None
    separable : bool, optional
        Whether to use separable convolutions, by default False
    complex_network : bool, optional
        Whether to use a complex network architecture, by default True
    complex_activation : str, optional
        Type of complex activation function to use, by default "real"
    spectral_layers : int, optional
        Number of spectral layers, by default 3
    checkpointing : int, optional
        Number of checkpointing segments, by default 0
    local_blocks: List[int], optional
        List of blocks to use local filters, by default []

    Example:
    --------
    >>> from modulus.models.sfno.sfnonet import SphericalFourierNeuralOperatorNet as SFNO
    >>> model = SFNO(
    ...         params={},
    ...         img_shape=(8, 16),
    ...         scale_factor=4,
    ...         in_chans=2,
    ...         out_chans=2,
    ...         embed_dim=16,
    ...         num_layers=2,
    ...         encoder_layers=1,
    ...         num_blocks=4,
    ...         spectral_layers=2,
    ...         use_mlp=True,)
    >>> model(torch.randn(1, 2, 8, 16)).shape
    torch.Size([1, 2, 8, 16])
    """

    def __init__(
        self,
        params,
        spectral_transform: str = "sht",
        filter_type: str = "linear",
        operator_type: str = "diagonal",
        img_shape: Tuple[int, int] = (721, 1440),
        scale_factor: int = 1,
        residual_filter_factor: int = 1,
        in_chans: int = 2,
        out_chans: int = 2,
        embed_dim: int = 256,
        context_config: ContextConfig = ContextConfig(
            embed_dim_scalar=0,
            embed_dim_2d=0,
        ),
        global_layer_norm: bool = False,
        num_layers: int = 12,
        use_mlp: int = True,
        mlp_ratio: float = 2.0,
        activation_function: str = "gelu",
        encoder_layers: int = 1,
        pos_embed: bool = True,
        drop_rate: float = 0.0,
        drop_path_rate: float = 0.0,
        num_blocks: int = 16,
        sparsity_threshold: float = 0.0,
        hard_thresholding_fraction: float = 1.0,
        use_complex_kernels: bool = True,
        big_skip: bool = True,
        rank: float = 1.0,
        factorization: Any = None,
        separable: bool = False,
        complex_network: bool = True,
        complex_activation: str = "real",
        spectral_layers: int = 3,
        checkpointing: int = 0,
        filter_residual: bool = False,
        filter_output: bool = False,
        local_blocks: Optional[List[int]] = None,
    ):
        super(SphericalFourierNeuralOperatorNet, self).__init__()

        self.params = params
        self.spectral_transform = (
            params.spectral_transform
            if hasattr(params, "spectral_transform")
            else spectral_transform
        )
        self.filter_type = (
            params.filter_type if hasattr(params, "filter_type") else filter_type
        )
        self.filter_residual = (
            params.filter_residual
            if hasattr(params, "filter_residual")
            else filter_residual
        )
        self.filter_output = (
            params.filter_output if hasattr(params, "filter_output") else filter_output
        )
        self.mlp_ratio = params.mlp_ratio if hasattr(params, "mlp_ratio") else mlp_ratio
        self.operator_type = (
            params.operator_type if hasattr(params, "operator_type") else operator_type
        )
        self.img_shape = (
            (params.img_shape_x, params.img_shape_y)
            if hasattr(params, "img_shape_x") and hasattr(params, "img_shape_y")
            else img_shape
        )
        self.scale_factor = (
            params.scale_factor if hasattr(params, "scale_factor") else scale_factor
        )
        self.residual_filter_factor = (
            params.residual_filter_factor
            if hasattr(params, "residual_filter_factor")
            else residual_filter_factor
        )
        if self.scale_factor != 1:
            raise NotImplementedError(
                "scale factor must be 1 as it is not implemented for "
                "conditional layer normalization"
            )
        self.global_layer_norm = (
            params.global_layer_norm
            if hasattr(params, "global_layer_norm")
            else global_layer_norm
        )
        self.in_chans = (
            params.N_in_channels if hasattr(params, "N_in_channels") else in_chans
        )
        self.out_chans = (
            params.N_out_channels if hasattr(params, "N_out_channels") else out_chans
        )
        self.embed_dim = self.num_features = (
            params.embed_dim if hasattr(params, "embed_dim") else embed_dim
        )
        self.num_layers = (
            params.num_layers if hasattr(params, "num_layers") else num_layers
        )
        self.num_blocks = (
            params.num_blocks if hasattr(params, "num_blocks") else num_blocks
        )
        self.hard_thresholding_fraction = (
            params.hard_thresholding_fraction
            if hasattr(params, "hard_thresholding_fraction")
            else hard_thresholding_fraction
        )
        self.use_mlp = params.use_mlp if hasattr(params, "use_mlp") else use_mlp
        self.activation_function = (
            params.activation_function
            if hasattr(params, "activation_function")
            else activation_function
        )
        self.encoder_layers = (
            params.encoder_layers
            if hasattr(params, "encoder_layers")
            else encoder_layers
        )
        self.pos_embed = params.pos_embed if hasattr(params, "pos_embed") else pos_embed
        self.big_skip = params.big_skip if hasattr(params, "big_skip") else big_skip
        self.rank = params.rank if hasattr(params, "rank") else rank
        self.factorization = (
            params.factorization if hasattr(params, "factorization") else factorization
        )
        self.separable = params.separable if hasattr(params, "separable") else separable
        self.complex_network = (
            params.complex_network
            if hasattr(params, "complex_network")
            else complex_network
        )
        self.complex_activation = (
            params.complex_activation
            if hasattr(params, "complex_activation")
            else complex_activation
        )
        self.spectral_layers = (
            params.spectral_layers
            if hasattr(params, "spectral_layers")
            else spectral_layers
        )
        self.checkpointing = (
            params.checkpointing if hasattr(params, "checkpointing") else checkpointing
        )
        local_blocks = (
            params.local_blocks if hasattr(params, "local_blocks") else local_blocks
        )
        if local_blocks is not None:
            self.local_blocks = [i for i in range(self.num_layers) if i in local_blocks]
        else:
            self.local_blocks = []

        data_grid = params.data_grid if hasattr(params, "data_grid") else "equiangular"
        # self.pretrain_encoding = params.pretrain_encoding if hasattr(params, "pretrain_encoding") else False

        # compute the downscaled image size
        self.h = int(self.img_shape[0] // self.scale_factor)
        self.w = int(self.img_shape[1] // self.scale_factor)

        # Compute the maximum frequencies in h and in w
        modes_lat = int(self.h * self.hard_thresholding_fraction)
        modes_lon = int((self.w // 2 + 1) * self.hard_thresholding_fraction)
        modes_lat_residual = int(self.img_shape[0] // self.residual_filter_factor)
        modes_lon_residual = int(
            self.img_shape[1] // self.residual_filter_factor // 2 + 1
        )

        # no global padding because we removed the horizontal distributed code
        self.padding = (0, 0)

        if residual_filter_factor == 1 and not self.filter_residual:
            self.residual_filter_down = nn.Identity()
            self.residual_filter_up = nn.Identity()
        else:
            self.residual_filter_down = th.RealSHT(
                *self.img_shape,
                lmax=modes_lat_residual,
                mmax=modes_lon_residual,
                grid=data_grid,
            ).float()
            self.residual_filter_up = th.InverseRealSHT(
                *self.img_shape,
                lmax=modes_lat_residual,
                mmax=modes_lon_residual,
                grid=data_grid,
            ).float()

        if self.filter_output:
            self.filter_output_down = th.RealSHT(
                *self.img_shape,
                lmax=modes_lat,
                mmax=modes_lon,
                grid=data_grid,
            ).float()
            self.filter_output_up = th.InverseRealSHT(
                *self.img_shape,
                lmax=modes_lat,
                mmax=modes_lon,
                grid=data_grid,
            ).float()
        else:
            self.filter_output_down = nn.Identity()
            self.filter_output_up = nn.Identity()

        # prepare the spectral transforms
        if self.spectral_transform == "sht":
            sht_handle = th.RealSHT
            isht_handle = th.InverseRealSHT

            # set up
            self.trans_down = sht_handle(
                *self.img_shape, lmax=modes_lat, mmax=modes_lon, grid=data_grid
            ).float()
            self.itrans_up = isht_handle(
                *self.img_shape, lmax=modes_lat, mmax=modes_lon, grid=data_grid
            ).float()
            self.trans = sht_handle(
                self.h, self.w, lmax=modes_lat, mmax=modes_lon, grid="legendre-gauss"
            ).float()
            self.itrans = isht_handle(
                self.h, self.w, lmax=modes_lat, mmax=modes_lon, grid="legendre-gauss"
            ).float()

        elif self.spectral_transform == "fft":
            fft_handle = th.RealFFT2
            ifft_handle = th.InverseRealFFT2

            # effective image size:
            self.img_shape_eff = (
                self.img_shape[0] + self.padding[0],
                self.img_shape[1] + self.padding[1],
            )
            self.img_shape_loc = (
                self.img_shape_eff[0],
                self.img_shape_eff[1],
            )

            self.trans_down = fft_handle(
                *self.img_shape_eff, lmax=modes_lat, mmax=modes_lon
            ).float()
            self.itrans_up = ifft_handle(
                *self.img_shape_eff, lmax=modes_lat, mmax=modes_lon
            ).float()
            self.trans = fft_handle(
                self.h, self.w, lmax=modes_lat, mmax=modes_lon
            ).float()
            self.itrans = ifft_handle(
                self.h, self.w, lmax=modes_lat, mmax=modes_lon
            ).float()
        else:
            raise (ValueError("Unknown spectral transform"))

        # use the SHT/FFT to compute the local, downscaled grid dimensions
        self.img_shape_loc = (self.trans_down.nlat, self.trans_down.nlon)
        self.img_shape_eff = (self.trans_down.nlat, self.trans_down.nlon)
        self.h_loc = self.itrans.nlat
        self.w_loc = self.itrans.nlon

        # determine activation function
        if self.activation_function == "relu":
            self.activation_function = nn.ReLU
        elif self.activation_function == "gelu":
            self.activation_function = nn.GELU
        elif self.activation_function == "silu":
            self.activation_function = nn.SiLU
        else:
            raise ValueError(f"Unknown activation function {self.activation_function}")

        # encoder
        encoder_hidden_dim = self.embed_dim
        current_dim = self.in_chans
        encoder_modules = []
        for i in range(self.encoder_layers):
            encoder_modules.append(
                nn.Conv2d(current_dim, encoder_hidden_dim, 1, bias=True)
            )
            encoder_modules.append(self.activation_function())
            current_dim = encoder_hidden_dim
        encoder_modules.append(nn.Conv2d(current_dim, self.embed_dim, 1, bias=False))
        self.encoder = nn.Sequential(*encoder_modules)

        # dropout
        self.pos_drop = nn.Dropout(p=drop_rate) if drop_rate > 0.0 else nn.Identity()
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, self.num_layers)]

        # FNO blocks
        self.blocks = nn.ModuleList([])
        for i in range(self.num_layers):
            if i in self.local_blocks:
                block_filter_type = "local"
            else:
                block_filter_type = self.filter_type

            first_layer = i == 0
            last_layer = i == self.num_layers - 1

            forward_transform = self.trans_down if first_layer else self.trans
            inverse_transform = self.itrans_up if last_layer else self.itrans

            inner_skip = "linear"
            outer_skip = "identity"

            block = FourierNeuralOperatorBlock(
                forward_transform,
                inverse_transform,
                self.embed_dim,
                context_config=context_config,
                filter_type=block_filter_type,
                operator_type=self.operator_type,
                mlp_ratio=self.mlp_ratio,
                drop_rate=drop_rate,
                drop_path=dpr[i],
                act_layer=self.activation_function,
                sparsity_threshold=sparsity_threshold,
                global_layer_norm=self.global_layer_norm,
                use_complex_kernels=use_complex_kernels,
                inner_skip=inner_skip,
                outer_skip=outer_skip,
                use_mlp=self.use_mlp,
                rank=self.rank,
                factorization=self.factorization,
                separable=self.separable,
                complex_network=self.complex_network,
                complex_activation=self.complex_activation,
                spectral_layers=self.spectral_layers,
                checkpointing=self.checkpointing,
                filter_residual=self.filter_residual,
            )

            self.blocks.append(block)

        # decoder
        decoder_hidden_dim = self.embed_dim
        current_dim = self.embed_dim + self.big_skip * self.in_chans
        decoder_modules = []
        for i in range(self.encoder_layers):
            decoder_modules.append(
                nn.Conv2d(current_dim, decoder_hidden_dim, 1, bias=True)
            )
            decoder_modules.append(self.activation_function())
            current_dim = decoder_hidden_dim
        decoder_modules.append(nn.Conv2d(current_dim, self.out_chans, 1, bias=False))
        self.decoder = nn.Sequential(*decoder_modules)

        # learned position embedding
        if self.pos_embed:
            # currently using deliberately a differently shape position embedding
            self.pos_embed = nn.Parameter(
                torch.zeros(
                    1, self.embed_dim, self.img_shape_loc[0], self.img_shape_loc[1]
                )
            )
            # self.pos_embed = nn.Parameter( torch.zeros(1, self.embed_dim, self.img_shape_eff[0], self.img_shape_eff[1]) )
            self.pos_embed.is_shared_mp = ["matmul"]
            trunc_normal_(self.pos_embed, std=0.02)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        """Helper routine for weight initialization"""
        if isinstance(m, nn.Linear) or isinstance(m, nn.Conv2d):
            trunc_normal_(m.weight, std=0.02)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, ConditionalLayerNorm):
            m.reset_parameters()

    @torch.jit.ignore
    def no_weight_decay(self):  # pragma: no cover
        """Helper"""
        return {"pos_embed", "cls_token"}

    def _forward_features(self, x: torch.Tensor, context: Context):
        for blk in self.blocks:
            if self.checkpointing >= 3:
                x = checkpoint(blk, x, context)
            else:
                x = blk(x, context)

        return x

    def forward(self, x: torch.Tensor, context: Context):
        # save big skip
        if self.big_skip:
            residual = self.residual_filter_up(self.residual_filter_down(x))

        if self.checkpointing >= 1:
            x = checkpoint(self.encoder, x)
        else:
            x = self.encoder(x)

        if hasattr(self, "pos_embed"):
            # old way of treating unequally shaped weights
            if self.img_shape_loc != self.img_shape_eff:
                xp = torch.zeros_like(x)
                xp[..., : self.img_shape_loc[0], : self.img_shape_loc[1]] = (
                    x[..., : self.img_shape_loc[0], : self.img_shape_loc[1]]
                    + self.pos_embed
                )
                x = xp
            else:
                x = x + self.pos_embed

        # maybe clean the padding just in case

        x = self.pos_drop(x)

        x = self._forward_features(x, context)

        if self.big_skip:
            x = torch.cat((x, residual), dim=1)

        if self.checkpointing >= 1:
            x = checkpoint(self.decoder, x)
        else:
            x = self.decoder(x)

        x = self.filter_output_up(self.filter_output_down(x))

        return x



import dataclasses
from collections.abc import Mapping
from typing import Any

import torch
from torch import nn

from .wildcard import apply_by_wildcard, wildcard_match


@dataclasses.dataclass
class CopyWeightsConfig:
    """
    Configuration for copying weights from a base model to a target model.

    Used during training to overwrite weights after every batch of data,
    to have the effect of "freezing" the overwritten weights. When the
    target parameters have longer dimensions than the base model, only
    the initial slice is overwritten.

    This is used to achieve an effect of freezing model parameters that
    can freeze a subset of each weight that comes from a smaller base weight.
    This is less efficient than true parameter freezing, but layer
    freezing is all-or-nothing for each parameter.

    All parameters must be covered by either the include or exclude list,
    but not both.

    Parameters:
        include: list of wildcard patterns to overwrite
        exclude: list of wildcard patterns to exclude from overwriting
    """

    include: list[str] = dataclasses.field(default_factory=list)
    exclude: list[str] = dataclasses.field(default_factory=list)

    def __post_init__(self):
        for pattern in self.include:
            if any(wildcard_match(pattern, exclude) for exclude in self.exclude):
                raise ValueError(
                    f"Parameter {pattern} is included in both include "
                    f"{self.include} and exclude {self.exclude}"
                )
        for pattern in self.exclude:
            if any(wildcard_match(pattern, include) for include in self.include):
                raise ValueError(
                    f"Parameter {pattern} is included in both include "
                    f"{self.include} and exclude {self.exclude}"
                )

    @torch.no_grad()
    def apply(self, weights: list[Mapping[str, Any]], modules: list[nn.Module]):
        """
        Apply base weights to modules according to the include/exclude lists
        of this instance.

        In order to "freeze" the weights during training, this must be called after
        each time the weights are updated in the training loop.

        Args:
            weights: list of base weights to apply
            modules: list of modules to apply the weights to
        """
        if len(modules) > 1:
            # We can support multiple modules by having this configuration take a list
            # of include/exclude for each module. Not implemented right now because it
            # is not needed, and would make the configuration more confusing for the
            # single-module case (especially when it's only ever single-module).
            raise NotImplementedError("only one module currently supported")
        if len(modules) != len(weights):
            raise ValueError("number of modules and weights must match")
        for module, weight in zip(modules, weights):

            def func(module, name):
                overwrite_weight_initial_slice(module, name, weight[name])

            apply_by_wildcard(module, func, self.include, self.exclude)
        return module


def strip_leading_module(state_dict: Mapping[str, Any]) -> Mapping[str, Any]:
    """
    Remove the leading "module." from the keys of a state dict.

    This is necessary because SingleModuleStepper wraps the module in either
    a DistributedDataParallel layer or DummyWrapper layer, which adds a leading
    "module." to the keys of the state dict.
    """
    return {
        k[len("module.") :] if k.startswith("module.") else k: v
        for k, v in state_dict.items()
    }


def overwrite_weights(
    from_state: Mapping[str, Any],
    to_module: torch.nn.Module,
    exclude_parameters: list[str] | None = None,
):
    """
    Overwrite the weights in to_module with the weights in from_state.

    When an axis is larger in to_module than in from_state, only the initial
    slice is overwritten. For example, if the from module has a parameter `a`
    of shape [10, 10], and the to module has a parameter `a` of shape [20, 10],
    then only the first 10 rows of `a` will be overwritten.

    If an axis is larger in from_state than in to_module, an exception is raised.

    Args:
        from_state: module state dict containing weights to be copied
        to_module: module whose weights will be overwritten
        exclude_parameters: list of parameter names to exclude from the loaded
            weights. Wildcards can be used, e.g. "decoder.*.weight".
    """
    if exclude_parameters is None:
        exclude_parameters = []
    from_names = set(from_state.keys())
    to_names = set(to_module.state_dict().keys())
    if not from_names.issubset(to_names):
        missing_parameters = from_names - to_names
        raise ValueError(
            f"Dest module is missing parameters {missing_parameters}, "
            "which is not allowed"
        )
    for name in from_names:
        if any(wildcard_match(pattern, name) for pattern in exclude_parameters):
            continue
        from_param = from_state[name]
        try:
            overwrite_weight_initial_slice(to_module, name, from_param)
        except AttributeError:  # if state is not a parameter
            pass


def overwrite_weight_initial_slice(module, name, from_param):
    """
    Overwrite the initial slice of a parameter in module with from_param.

    When an axis is larger in the module's param than in from_param,
    only the initial slice is overwritten. For example, if the from module
    has a parameter `a` of shape [10, 10], and the to module has a parameter
    `a` of shape [20, 10], then only the first 10 rows of `a` will be overwritten.

    If an axis is larger in from_param, an exception is raised.

    Args:
        module: module whose parameter will be overwritten
        name: name of the parameter to be overwritten
        from_param: parameter to overwrite with
    """
    try:
        to_param = module.get_parameter(name)
    except AttributeError:
        if name == "device_buffer" or name == "module.device_buffer":
            return  # ignore device buffer, used for GPU operations
        raise
    if len(from_param.shape) != len(to_param.shape):
        raise ValueError(
            f"Dest parameter {name} has "
            f"{len(to_param.shape.shape)} "
            "dimensions which needs to be equal to the loaded "
            f"parameter dimension {len(from_param.shape)}"
        )
    for from_size, to_size in zip(from_param.shape, to_param.shape):
        if from_size > to_size:
            raise ValueError(
                f"Dest parameter has size {to_size} along one of its "
                "dimensions which needs to be greater than loaded "
                f"parameter size {from_size}"
            )
    slices = tuple(slice(0, size) for size in from_param.shape)
    with torch.no_grad():
        new_param_data = to_param.data.clone()
        new_param_data[slices] = from_param.data
        _set_nested_parameter(module, name, new_param_data)


def _set_nested_parameter(module, param_name, new_param):
    *path, name = param_name.split(".")
    for p in path:
        module = getattr(module, p)
    getattr(module, name)[:] = new_param



from collections.abc import Mapping
from typing import Protocol

import torch

from fme.core import metrics
from fme.core.constants import (
    GRAVITY,
    LATENT_HEAT_OF_VAPORIZATION,
    RDGAS,
    RVGAS,
    SPECIFIC_HEAT_OF_DRY_AIR_CONST_PRESSURE,
    SPECIFIC_HEAT_OF_DRY_AIR_CONST_VOLUME,
)
from fme.core.stacker import Stacker
from fme.core.typing_ import TensorDict, TensorMapping

ATMOSPHERE_FIELD_NAME_PREFIXES = {
    "specific_total_water": ["specific_total_water_"],
    "surface_pressure": ["PRESsfc", "PS"],
    "surface_height": ["HGTsfc"],
    "surface_geopotential": ["PHIS"],
    "tendency_of_total_water_path_due_to_advection": [
        "tendency_of_total_water_path_due_to_advection"
    ],
    "latent_heat_flux": ["LHTFLsfc", "LHFLX"],
    "sensible_heat_flux": ["SHTFLsfc", "SHFLX"],
    "precipitation_rate": ["PRATEsfc", "surface_precipitation_rate"],
    "sfc_down_sw_radiative_flux": ["DSWRFsfc", "FSDS"],
    "sfc_up_sw_radiative_flux": ["USWRFsfc", "surface_upward_shortwave_flux"],
    "sfc_down_lw_radiative_flux": ["DLWRFsfc", "FLDS"],
    "sfc_up_lw_radiative_flux": ["ULWRFsfc", "surface_upward_longwave_flux"],
    "toa_up_lw_radiative_flux": ["ULWRFtoa", "FLUT"],
    "toa_up_sw_radiative_flux": ["USWRFtoa", "top_of_atmos_upward_shortwave_flux"],
    "toa_down_sw_radiative_flux": ["DSWRFtoa", "SOLIN"],
    "air_temperature": ["air_temperature_", "T_"],
    "frozen_precipitation_rate": ["total_frozen_precipitation_rate"],
    "eastward_wind_at_10m": ["UGRD10m"],
    "northward_wind_at_10m": ["VGRD10m"],
}


class HasAtmosphereVerticalIntegral(Protocol):
    def vertical_integral(
        self,
        field: torch.Tensor,
        surface_pressure: torch.Tensor,
    ) -> torch.Tensor:
        pass

    def interface_pressure(self, surface_pressure: torch.Tensor) -> torch.Tensor:
        pass

    def get_ak(self) -> torch.Tensor:
        pass

    def get_bk(self) -> torch.Tensor:
        pass


class AtmosphereData:
    """Container for atmospheric data for accessing variables and providing
    torch.Tensor views on data with multiple vertical levels.
    """

    def __init__(
        self,
        atmosphere_data: TensorMapping,
        vertical_coordinate: HasAtmosphereVerticalIntegral | None = None,
        atmosphere_field_name_prefixes: Mapping[str, list[str]] | None = None,
    ):
        """
        Initializes the instance based on the provided data and prefixes.

        Args:
            atmosphere_data: Mapping from field names to tensors.
            vertical_coordinate: The vertical coordinate of the model. If not provided,
                then variables which require vertical integration will raise an error.
            atmosphere_field_name_prefixes: Mapping which defines the correspondence
                between an arbitrary set of "standard" names (e.g., "surface_pressure"
                or "air_temperature") and lists of possible names or prefix variants
                (e.g., ["PRESsfc", "PS"] or ["air_temperature_", "T_"]) found in the
                data.
        """
        if atmosphere_field_name_prefixes is None:
            atmosphere_field_name_prefixes = ATMOSPHERE_FIELD_NAME_PREFIXES.copy()
        self._data = dict(atmosphere_data)
        self._prefix_map = atmosphere_field_name_prefixes
        self._vertical_coordinate = vertical_coordinate
        self._stacker = Stacker(atmosphere_field_name_prefixes)

    @property
    def data(self) -> TensorDict:
        """Mapping from field names to tensors."""
        return self._data

    def __getitem__(self, name: str):
        return getattr(self, name)

    def _get_prefix(self, prefix):
        return self.data[prefix]

    def _set(self, name, value):
        for prefix in self._prefix_map[name]:
            if prefix in self.data.keys():
                self._set_prefix(prefix, value)
                return
        raise KeyError(name)

    def _set_prefix(self, prefix, value):
        self.data[prefix] = value

    def _get(self, name):
        for prefix in self._prefix_map[name]:
            if prefix in self.data.keys():
                return self._get_prefix(prefix)
        raise KeyError(name)

    @property
    def air_temperature(self) -> torch.Tensor:
        """Returns all vertical levels of air_temperature, e.g. a tensor of
        shape `(..., vertical_level)`.
        """
        return self._stacker("air_temperature", self.data)

    @property
    def specific_total_water(self) -> torch.Tensor:
        """Returns all vertical levels of specific total water, e.g. a tensor of
        shape `(..., vertical_level)`.
        """
        return self._stacker("specific_total_water", self.data)

    @property
    def surface_height(self) -> torch.Tensor:
        try:
            return self._get("surface_height")
        except KeyError:
            # E3SM saves geopotential not surface height so need to convert
            # by using g value from e3sm
            GRAVITY_E3SM = 9.80616
            return self._get("surface_geopotential") / GRAVITY_E3SM

    @property
    def surface_pressure(self) -> torch.Tensor:
        return self._get("surface_pressure")

    def set_surface_pressure(self, value: torch.Tensor):
        self._set("surface_pressure", value)

    @property
    def toa_down_sw_radiative_flux(self) -> torch.Tensor:
        return self._get("toa_down_sw_radiative_flux")

    @property
    def toa_up_sw_radiative_flux(self) -> torch.Tensor:
        return self._get("toa_up_sw_radiative_flux")

    @property
    def toa_up_lw_radiative_flux(self) -> torch.Tensor:
        return self._get("toa_up_lw_radiative_flux")

    @property
    def surface_pressure_due_to_dry_air(self) -> torch.Tensor:
        if self._vertical_coordinate is None:
            raise ValueError("Vertical coordinate must be provided to compute dry air.")
        return metrics.surface_pressure_due_to_dry_air(
            self.surface_pressure,
            self.total_water_path,
        )

    @property
    def total_water_path(self) -> torch.Tensor:
        if self._vertical_coordinate is None:
            raise ValueError(
                "Vertical coordinate must be provided to compute total water path."
            )
        return self._vertical_coordinate.vertical_integral(
            self.specific_total_water,
            self.surface_pressure,
        )

    @property
    def frozen_precipitation_rate(self) -> torch.Tensor:
        # Return zero if any necessary fields are missing
        try:
            return self._get("frozen_precipitation_rate")
        except KeyError:
            try:
                return (
                    self._data["ICEsfc"]
                    + self._data["GRAUPELsfc"]
                    + self._data["SNOWsfc"]
                )
            except KeyError:
                return torch.zeros_like(self.surface_pressure)

    @property
    def net_surface_energy_flux_without_frozen_precip(self) -> torch.Tensor:
        return metrics.net_surface_energy_flux(
            self._get("sfc_down_lw_radiative_flux"),
            self._get("sfc_up_lw_radiative_flux"),
            self._get("sfc_down_sw_radiative_flux"),
            self._get("sfc_up_sw_radiative_flux"),
            self._get("latent_heat_flux"),
            self._get("sensible_heat_flux"),
        )

    @property
    def net_surface_energy_flux(self) -> torch.Tensor:
        return metrics.net_surface_energy_flux(
            self._get("sfc_down_lw_radiative_flux"),
            self._get("sfc_up_lw_radiative_flux"),
            self._get("sfc_down_sw_radiative_flux"),
            self._get("sfc_up_sw_radiative_flux"),
            self._get("latent_heat_flux"),
            self._get("sensible_heat_flux"),
            frozen_precipitation_rate=self.frozen_precipitation_rate,
        )

    @property
    def net_top_of_atmosphere_energy_flux(self) -> torch.Tensor:
        return metrics.net_top_of_atmosphere_energy_flux(
            self._get("toa_down_sw_radiative_flux"),
            self._get("toa_up_sw_radiative_flux"),
            self._get("toa_up_lw_radiative_flux"),
        )

    @property
    def net_energy_flux_into_atmosphere(self) -> torch.Tensor:
        return self.net_top_of_atmosphere_energy_flux - self.net_surface_energy_flux

    @property
    def precipitation_rate(self) -> torch.Tensor:
        """
        Precipitation rate in kg m-2 s-1.
        """
        return self._get("precipitation_rate")

    def set_precipitation_rate(self, value: torch.Tensor):
        self._set("precipitation_rate", value)

    @property
    def latent_heat_flux(self) -> torch.Tensor:
        """
        Latent heat flux in W m-2.
        """
        return self._get("latent_heat_flux")

    @property
    def evaporation_rate(self) -> torch.Tensor:
        """
        Evaporation rate in kg m-2 s-1.
        """
        lhf = self._get("latent_heat_flux")  # W/m^2
        # (W/m^2) / (J/kg) = (J s^-1 m^-2) / (J/kg) = kg/m^2/s
        return lhf / LATENT_HEAT_OF_VAPORIZATION

    def set_evaporation_rate(self, value: torch.Tensor):
        self._set("latent_heat_flux", value * LATENT_HEAT_OF_VAPORIZATION)

    @property
    def tendency_of_total_water_path_due_to_advection(self) -> torch.Tensor:
        """
        Tendency of total water path due to advection in kg m-2 s-1.
        """
        return self._get("tendency_of_total_water_path_due_to_advection")

    def set_tendency_of_total_water_path_due_to_advection(self, value: torch.Tensor):
        self._set("tendency_of_total_water_path_due_to_advection", value)

    def height_at_log_midpoint(self) -> torch.Tensor:
        """
        Compute vertical height at layer log midpoints.
        """
        if self._vertical_coordinate is None:
            raise ValueError(
                "Vertical coordinate must be provided to compute height at log midpoint"
            )
        interface_pressure = self._vertical_coordinate.interface_pressure(
            self.surface_pressure
        )
        layer_thickness = compute_layer_thickness(
            pressure_at_interface=interface_pressure,
            air_temperature=self.air_temperature,
            specific_total_water=self.specific_total_water,
        )
        height_at_interface = _height_at_interface(layer_thickness, self.surface_height)
        return (height_at_interface[..., :-1] * height_at_interface[..., 1:]) ** 0.5

    @property
    def height_at_midpoint(self) -> torch.Tensor:
        """Compute vertical height at layer midpoints with linear interpolation."""
        if self._vertical_coordinate is None:
            raise ValueError(
                "Vertical coordinate must be provided to compute height at mmidpoint"
            )
        interface_pressure = self._vertical_coordinate.interface_pressure(
            self.surface_pressure
        )
        layer_thickness = compute_layer_thickness(
            pressure_at_interface=interface_pressure,
            air_temperature=self.air_temperature,
            specific_total_water=self.specific_total_water,
        )
        height_at_interface = _height_at_interface(layer_thickness, self.surface_height)
        return 0.5 * (height_at_interface[..., :-1] + height_at_interface[..., 1:])

    @property
    def moist_static_energy(self) -> torch.Tensor:
        """
        Compute moist static energy.
        """
        # ACE does not currently prognose specific humidity, so here we closely
        # approximate this using specific total water (<0.01% effect on total MSE).
        return (
            self.air_temperature * SPECIFIC_HEAT_OF_DRY_AIR_CONST_PRESSURE
            + self.specific_total_water * LATENT_HEAT_OF_VAPORIZATION
            + self.height_at_midpoint * GRAVITY
        )

    @property
    def total_energy_ace2(self) -> torch.Tensor:
        """
        Compute the total energy, following some assumptions used for ACE2 models.

        Namely, we ignore kinetic energy, use hydrostatic balance to compute the
        geoportential energy, and approximate specific humidity with specific total
        water. We also ignore the ice water contribution to total energy.
        """
        return (
            self.air_temperature * SPECIFIC_HEAT_OF_DRY_AIR_CONST_VOLUME
            + self.specific_total_water * LATENT_HEAT_OF_VAPORIZATION
            + self.height_at_midpoint * GRAVITY
        )

    @property
    def total_energy_ace2_path(self) -> torch.Tensor:
        """Compute vertical integral of total energy."""
        if self._vertical_coordinate is None:
            raise ValueError(
                "Vertical coordinate must be provided to compute total energy ACE2 path"
            )
        return self._vertical_coordinate.vertical_integral(
            self.total_energy_ace2, self.surface_pressure
        )

    @property
    def windspeed_at_10m(self) -> torch.Tensor:
        """Compute the windspeed at 10m above surface."""
        return torch.sqrt(
            self._get("eastward_wind_at_10m") ** 2
            + self._get("northward_wind_at_10m") ** 2
        )


def compute_layer_thickness(
    pressure_at_interface: torch.Tensor,
    air_temperature: torch.Tensor,
    specific_total_water: torch.Tensor,
) -> torch.Tensor:
    """
    Computes vertical thickness of each layer assuming hydrostatic equilibrium.
    ACE does not currently prognose specific humidity, so here we closely
    approximate this using specific total water.
    """
    tv = air_temperature * (1 + (RVGAS / RDGAS - 1.0) * specific_total_water)
    # Enforce min log(p) = 0 so that geopotential energy calculation is finite.
    # This is equivalent to setting the TOA pressure to 1 Pa if it is less than that.
    # The ERA5 data has a TOA pressure of 0.0 Pa which causes issues otherwise.
    dlogp = torch.clamp(torch.log(pressure_at_interface), min=0.0).diff(dim=-1)
    return dlogp * RDGAS * tv / GRAVITY


def _height_at_interface(
    layer_thickness: torch.Tensor, surface_height: torch.Tensor
) -> torch.Tensor:
    """
    Computes height at layer interfaces from layer thickness and surface height.
    Vertical coordinate is the last tensor dimension.
    """
    cumulative_thickness = torch.cumsum(layer_thickness.flip(dims=(-1,)), dim=-1).flip(
        dims=(-1,)
    )
    # Sometimes surface height data has negative values, which are filled with 0.
    hsfc = torch.where(surface_height < 0.0, 0, surface_height).reshape(
        *surface_height.shape, 1
    )
    return torch.concat(
        [
            (cumulative_thickness + hsfc.broadcast_to(cumulative_thickness.shape)),
            hsfc,
        ],
        dim=-1,
    )



# flake8: noqa
# Copied from https://github.com/ai2cm/modulus/commit/22df4a9427f5f12ff6ac891083220e7f2f54d229
# Copyright (c) 2023, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import warnings

import torch


def _no_grad_trunc_normal_(tensor, mean, std, a, b):
    # Cut & paste from PyTorch official master until it's in a few official releases - RW
    # Method based on https://people.sc.fsu.edu/~jburkardt/presentations/truncated_normal.pdf
    def norm_cdf(x):
        # Computes standard normal cumulative distribution function
        return (1.0 + math.erf(x / math.sqrt(2.0))) / 2.0

    if (mean < a - 2 * std) or (mean > b + 2 * std):
        warnings.warn(
            "mean is more than 2 std from [a, b] in nn.init.trunc_normal_. "
            "The distribution of values may be incorrect.",
            stacklevel=2,
        )

    with torch.no_grad():
        # Values are generated by using a truncated uniform distribution and
        # then using the inverse CDF for the normal distribution.
        # Get upper and lower cdf values
        l = norm_cdf((a - mean) / std)
        u = norm_cdf((b - mean) / std)

        # Uniformly fill tensor with values from [l, u], then translate to
        # [2l-1, 2u-1].
        tensor.uniform_(2 * l - 1, 2 * u - 1)

        # Use inverse cdf transform for normal distribution to get truncated
        # standard normal
        tensor.erfinv_()

        # Transform to proper mean, std
        tensor.mul_(std * math.sqrt(2.0))
        tensor.add_(mean)

        # Clamp to ensure it's in the proper range
        tensor.clamp_(min=a, max=b)
        return tensor


def trunc_normal_(tensor, mean=0.0, std=1.0, a=-2.0, b=2.0):
    r"""Fills the input Tensor with values drawn from a truncated
    normal distribution. The values are effectively drawn from the
    normal distribution :math:`\mathcal{N}(\text{mean}, \text{std}^2)`
    with values outside :math:`[a, b]` redrawn until they are within
    the bounds. The method used for generating the random values works
    best when :math:`a \leq \text{mean} \leq b`.
    Args:
    tensor: an n-dimensional `torch.Tensor`
    mean: the mean of the normal distribution
    std: the standard deviation of the normal distribution
    a: the minimum cutoff value
    b: the maximum cutoff value
    """
    return _no_grad_trunc_normal_(tensor, mean, std, a, b)



import datetime

import pytest
import torch

from fme.core.atmosphere_data import AtmosphereData
from fme.core.coordinates import HybridSigmaPressureCoordinate
from fme.core.derived_variables import (
    _compute_derived_variable,
    compute_derived_quantities,
    get_derived_variable_metadata,
)
from fme.core.typing_ import TensorDict, TensorMapping

TIMESTEP = datetime.timedelta(hours=6)


def test_compute_derived_variable():
    fake_data = {"PRESsfc": torch.tensor([1.0]), "PRATEsfc": torch.tensor([2.0])}
    vertical_coordinate = HybridSigmaPressureCoordinate(
        ak=torch.tensor([0.0, 0.0]), bk=torch.tensor([0.0, 1.0])
    )

    def _derived_variable_func(data: AtmosphereData, *_) -> torch.Tensor:
        return data.surface_pressure + data.precipitation_rate

    output_data = _compute_derived_variable(
        fake_data, vertical_coordinate, TIMESTEP, "c", _derived_variable_func
    )
    torch.testing.assert_close(output_data["c"], torch.tensor([3.0]))


def test_compute_derived_variable_raises_value_error_when_overwriting():
    fake_data = {"PRESsfc": torch.tensor([1.0]), "PRATEsfc": torch.tensor([2.0])}
    vertical_coordinate = HybridSigmaPressureCoordinate(
        ak=torch.tensor([0.0, 0.0]), bk=torch.tensor([0.0, 1.0])
    )

    def add_surface_pressure_and_precipitation(
        data: AtmosphereData, *_
    ) -> torch.Tensor:
        return data.surface_pressure + data.precipitation_rate

    derived_variable_func = add_surface_pressure_and_precipitation
    with pytest.raises(ValueError):
        _compute_derived_variable(
            fake_data, vertical_coordinate, TIMESTEP, "PRATEsfc", derived_variable_func
        )


@pytest.mark.parametrize("dataset", ["fv3", "e3sm"])
def test_compute_derived_quantities(dataset: str):
    torch.manual_seed(0)

    if dataset == "fv3":
        fake_data = {
            "PRESsfc": 10.0 + torch.rand(2, 3, 4, 8),
            "specific_total_water_0": torch.rand(2, 3, 4, 8),
            "specific_total_water_1": torch.rand(2, 3, 4, 8),
            "PRATEsfc": torch.rand(2, 3, 4, 8),
            "LHTFLsfc": torch.rand(2, 3, 4, 8),
            "tendency_of_total_water_path_due_to_advection": torch.rand(2, 3, 4, 8),
            "DSWRFtoa": torch.rand(2, 3, 4, 8),
            "USWRFtoa": torch.rand(2, 3, 4, 8),
            "ULWRFtoa": torch.rand(2, 3, 4, 8),
        }
        gen_data = fake_data.copy()
        del gen_data["DSWRFtoa"]

    if dataset == "e3sm":
        fake_data = {
            "PS": 10.0 + torch.rand(2, 3, 4, 8),
            "specific_total_water_0": torch.rand(2, 3, 4, 8),
            "specific_total_water_1": torch.rand(2, 3, 4, 8),
            "surface_precipitation_rate": torch.rand(2, 3, 4, 8),
            "LHFLX": torch.rand(2, 3, 4, 8),
            "tendency_of_total_water_path_due_to_advection": torch.rand(2, 3, 4, 8),
            "SOLIN": torch.rand(2, 3, 4, 8),
            "top_of_atmos_upward_shortwave_flux": torch.rand(2, 3, 4, 8),
            "FLUT": torch.rand(2, 3, 4, 8),
        }
        gen_data = fake_data.copy()
        del gen_data["SOLIN"]

    vertical_coordinate = HybridSigmaPressureCoordinate(
        ak=torch.tensor([0.0, 0.5, 0.0]),
        bk=torch.tensor([0.0, 0.5, 1.0]),
    )

    def derive_func(data: TensorMapping, forcing_data: TensorMapping) -> TensorDict:
        updated = compute_derived_quantities(
            dict(data),
            vertical_coordinate=vertical_coordinate,
            timestep=TIMESTEP,
            forcing_data=dict(forcing_data),
        )
        return updated

    out_gen_data = derive_func(gen_data, fake_data)
    out_target_data = derive_func(fake_data, fake_data)
    for name in (
        "total_water_path_budget_residual",
        "total_water_path",
        "surface_pressure_due_to_dry_air",
        "surface_pressure_due_to_dry_air_absolute_tendency",
        "net_energy_flux_toa_into_atmosphere",
    ):
        assert name in out_gen_data
        assert name in out_target_data
        assert out_gen_data[name].shape == (2, 3, 4, 8)
        assert out_target_data[name].shape == (2, 3, 4, 8)


def test_metadata_registry():
    metadata = get_derived_variable_metadata()
    assert metadata["total_water_path"].units == "kg/m**2"
    assert metadata["total_water_path"].long_name == "Total water path"



import numpy as np
import pytest
import torch

from fme.ace.aggregator.one_step.reduced import MeanAggregator
from fme.core.device import get_device
from fme.core.gridded_ops import LatLonOperations
from fme.core.testing import mock_distributed


def test_mean_metrics_call_distributed():
    """
    All reduced metrics should be reduced across processes using Distributed.

    This tests that functionality by modifying the Distributed singleton.
    """
    with mock_distributed(-1.0) as mock:
        area_weights = torch.ones([4]).to(get_device())
        agg = MeanAggregator(LatLonOperations(area_weights))
        sample_data = {"a": torch.ones([2, 3, 4, 4], device=get_device())}
        agg.record_batch(
            loss=1.0,
            target_data=sample_data,
            gen_data=sample_data,
            target_data_norm=sample_data,
            gen_data_norm=sample_data,
        )
        logs = agg.get_logs(label="metrics")
        assert logs["metrics/loss"] == -1.0
        assert logs["metrics/weighted_rmse/a"] == -1.0
        assert logs["metrics/weighted_grad_mag_percent_diff/a"] == -1.0
        assert mock.reduce_called


def test_i_time_start_gets_correct_time_one_step_windows():
    # while this directly tests the "mean" result, this is really a test that
    # the data from the correct timestep is piped into the aggregator.
    target_time = 3
    area_weights = torch.ones([4]).to(get_device())
    agg = MeanAggregator(LatLonOperations(area_weights), target_time=target_time)
    target_data = {"a": torch.zeros([2, 1, 4, 4], device=get_device())}
    for i in range(5):
        sample_data = {
            "a": torch.full([2, 1, 4, 4], fill_value=float(i), device=get_device())
        }
        agg.record_batch(
            loss=1.0,
            target_data=target_data,
            gen_data=sample_data,
            target_data_norm=target_data,
            gen_data_norm=sample_data,
            i_time_start=i,
        )
    logs = agg.get_logs(label="metrics")
    np.testing.assert_allclose(logs["metrics/weighted_bias/a"], target_time)


@pytest.mark.parametrize(
    "window_len, n_windows, target_time",
    [
        pytest.param(3, 1, 1, id="single_window"),
        pytest.param(3, 2, 1, id="first_of_two_windows"),
        pytest.param(3, 2, 4, id="second_of_two_windows"),
    ],
)
def test_i_time_start_gets_correct_time_longer_windows(
    window_len: int, n_windows: int, target_time: int
):
    # while this directly tests the "mean" result, this is really a test that
    # the data from the correct timestep is piped into the aggregator.
    area_weights = torch.ones([4]).to(get_device())
    agg = MeanAggregator(LatLonOperations(area_weights), target_time=target_time)
    target_data = {"a": torch.zeros([2, window_len, 4, 4], device=get_device())}
    i_start = 0
    for i in range(n_windows):
        sample_data = {"a": torch.zeros([2, window_len, 4, 4], device=get_device())}
        for i in range(window_len):
            sample_data["a"][..., i, :, :] = float(i_start + i)
        agg.record_batch(
            loss=1.0,
            target_data=target_data,
            gen_data=sample_data,
            target_data_norm=target_data,
            gen_data_norm=sample_data,
            i_time_start=i_start,
        )
        i_start += window_len
    logs = agg.get_logs(label="metrics")
    np.testing.assert_allclose(
        float(logs["metrics/weighted_bias/a"]), float(target_time), rtol=1e-5
    )


def test_loss():
    """
    Basic test the aggregator combines loss correctly
    with multiple batches and no distributed training.
    """
    torch.manual_seed(0)
    example_data = {
        "a": torch.randn(1, 2, 5, 5, device=get_device()),
    }
    area_weights = torch.ones(1).to(get_device())
    aggregator = MeanAggregator(LatLonOperations(area_weights))
    aggregator.record_batch(
        loss=1.0,
        target_data=example_data,
        gen_data=example_data,
        target_data_norm=example_data,
        gen_data_norm=example_data,
    )
    aggregator.record_batch(
        loss=2.0,
        target_data=example_data,
        gen_data=example_data,
        target_data_norm=example_data,
        gen_data_norm=example_data,
    )
    logs = aggregator.get_logs(label="metrics")
    assert logs["metrics/loss"] == 1.5
    aggregator.record_batch(
        loss=3.0,
        target_data=example_data,
        gen_data=example_data,
        target_data_norm=example_data,
        gen_data_norm=example_data,
    )
    logs = aggregator.get_logs(label="metrics")
    assert logs["metrics/loss"] == 2.0



from typing import Any

import torch

from .typing_ import EnsembleTensorDict, TensorDict, TensorMapping


def assert_dict_allclose(a: dict[str, Any], b: dict[str, Any]):
    """
    Check if two (possibly nested) dictionaries which may contain tensors are close.

    Non-tensor values are checked for equality.
    """
    if a.keys() != b.keys():
        raise AssertionError(f"Keys do not match, got {a.keys()} and {b.keys()}")
    for k in a.keys():
        if isinstance(a[k], torch.Tensor) and isinstance(b[k], torch.Tensor):
            torch.testing.assert_close(
                a[k], b[k], msg=f"Tensors for key {k} are not close"
            )
        elif isinstance(a[k], dict) and isinstance(b[k], dict):
            assert_dict_allclose(a[k], b[k])
        elif a[k] != b[k]:
            raise AssertionError(f"Values for key {k} are not equal")


def add_ensemble_dim(d: TensorMapping, repeats: int = 1) -> EnsembleTensorDict:
    """
    Add an explicit ensemble dimension to a tensor dict.

    Args:
        d: The tensor dict to add the ensemble dimension to.
        repeats: The number of ensemble members. If greater than 1, the ensemble
            will consist of repeated copies of the input tensor.

    Returns:
        The tensor dict with an explicit ensemble dimension.
    """
    return EnsembleTensorDict(
        {
            k: v[:, None, ...].repeat(1, repeats, *([1] * (v.ndim - 1)))
            for k, v in d.items()
        }
    )


def fold_sized_ensemble_dim(d: EnsembleTensorDict, n_ensemble: int) -> TensorDict:
    """
    Take a tensor dict with an explicit ensemble dimension and fold it into a
    batch/sample dimension.

    If the sample dimension is 1, it may be broadcasted to a larger n_ensemble
    count before folding.

    Args:
        d: The tensor dict to fold.
        n_ensemble: The number of ensemble members.

    Returns:
        The folded tensor dict.
    """
    for v in d.values():
        input_n_ensemble = v.shape[1]
        break
    else:
        return d  # empty data
    if input_n_ensemble == 1 and n_ensemble != 1:
        d = EnsembleTensorDict(
            {k: v.repeat(1, n_ensemble, *([1] * (v.ndim - 2))) for k, v in d.items()}
        )
    elif input_n_ensemble != n_ensemble:
        raise ValueError(
            f"Input n_ensemble ({input_n_ensemble}) does not match "
            f"n_ensemble ({n_ensemble}) in given data"
        )
    try:
        reshaped = {
            k: v.reshape(v.shape[0] * n_ensemble, *v.shape[2:]) for k, v in d.items()
        }
    except RuntimeError as e:
        if "is invalid for input of size" in str(e):
            raise ValueError(
                f"some values in d have invalid ensemble member counts, "
                f"should be {n_ensemble}, got: "
                f"{', '.join(f'{k}: {str(v.shape[1])}' for k, v in d.items())}"
            ) from e
        raise
    return EnsembleTensorDict(reshaped)


def fold_ensemble_dim(d: EnsembleTensorDict) -> tuple[TensorDict, int]:
    """
    Take a tensor dict with an explicit ensemble dimension and fold it into a
    batch/sample dimension.

    Args:
        d: The ensemble tensor dict to fold.

    Returns:
        A tuple of the folded tensor dict and the ensemble dimension length.
    """
    for v in d.values():
        n_ensemble = v.shape[1]
        break
    else:
        raise ValueError("input is empty, ensemble member count is not defined")
    return fold_sized_ensemble_dim(d, n_ensemble), n_ensemble


def repeat_interleave_batch_dim(data: TensorMapping, repeats: int) -> TensorDict:
    """
    Repeats each sample in the batch a given number of times using repeat_interleave.

    If you were to "unfold" the result of this operation with the repeat count as
    the ensemble dimension, you would get an ensemble tensor dict with the data
    repeated across the ensemble dimension.
    """
    if repeats == 1:
        return dict(data)  # no-op
    return {k: v.repeat_interleave(repeats, dim=0) for k, v in data.items()}


def unfold_ensemble_dim(d: TensorDict, n_ensemble: int) -> EnsembleTensorDict:
    """
    Take a tensor dict with a batch/sample dimension and unfold it into an
    explicit ensemble dimension.

    If an ensemble tensor dict is folded, this will unfold it back to the
    original ensemble dimension.

    Args:
        d: The tensor dict to unfold.
        n_ensemble: The number of ensemble members.

    Returns:
        The unfolded ensemble tensor dict.
    """
    return EnsembleTensorDict(
        {
            k: v.reshape(v.shape[0] // n_ensemble, n_ensemble, *v.shape[1:])
            for k, v in d.items()
        }
    )



import pathlib
from typing import List

import pytest
import xarray as xr
from write_monthly_data import Config, run

from fme.ace.data_loading.config import ConcatDatasetConfig, DataLoaderConfig
from fme.ace.testing import DimSize, DimSizes
from fme.ace.testing.fv3gfs_data import save_nd_netcdf
from fme.core.dataset.config import XarrayDataConfig
from fme.core.logging_utils import LoggingConfig


def write_ensemble_dataset(
    path: pathlib.Path, n_members: int, names: List[str], dim_sizes: DimSizes
):
    if not path.exists():
        path.mkdir(parents=True)
    for i in range(n_members):
        ensemble_dir = path / f"ic_{i:04d}"
        ensemble_dir.mkdir(exist_ok=True)
        save_nd_netcdf(
            ensemble_dir / "data.nc",
            dim_sizes,
            names,
            timestep_days=5,
        )


def test_write_monthly_data(very_fast_only: bool, tmp_path: pathlib.Path):
    if very_fast_only:
        pytest.skip("Skipping non-fast tests")
    all_names = ["a", "b"]
    horizontal = [DimSize("lat", 8), DimSize("lon", 4)]
    dim_sizes = DimSizes(
        n_time=4 * 60,
        horizontal=horizontal,
        nz_interface=2,
    )
    n_members = 3
    write_ensemble_dataset(tmp_path / "data", n_members, all_names, dim_sizes)
    dataset = [
        XarrayDataConfig(data_path=str(tmp_path / "data" / f"ic_{i:04}"))
        for i in range(n_members)
    ]
    config = Config(
        experiment_dir=str(tmp_path),
        data_loader=DataLoaderConfig(
            dataset=ConcatDatasetConfig(concat=dataset),
            batch_size=1,
            num_data_workers=0,
        ),
        logging=LoggingConfig(
            log_to_screen=True,
            log_to_file=False,
            log_to_wandb=False,
        ),
        variable_names=all_names,
    )
    run(config)
    xr.open_dataset(tmp_path / "monthly_mean_data.nc", decode_timedelta=False)



import numpy as np
import torch

import fme
from fme.ace.aggregator.inference.reduced import (
    AreaWeightedReducedMetric,
    SingleTargetMeanAggregator,
)
from fme.core.device import get_device
from fme.core.gridded_ops import LatLonOperations
from fme.core.typing_ import TensorDict, TensorMapping


def test_area_weighted_reduced_metric_includes_later_window_starts():
    """
    The area weighted reduced metric should assume that the start
    of a window is always recorded, as we clip it before calling.
    """

    def compute_metric(
        truth: TensorMapping, predicted: TensorMapping, weights=None, dim=()
    ) -> TensorDict:
        return {name: truth[name].mean(dim=(2, 3)) for name in truth}

    metric = AreaWeightedReducedMetric(
        device=get_device(),
        compute_metric=compute_metric,
        n_timesteps=7,
    )

    data = torch.ones([2, 3, 4, 4], device=get_device())
    data_dict = {"var1": data}
    metric.record(data_dict, data_dict, 0)
    data_nan = data.clone()
    data_nan[:, 0, :, :] = np.nan
    data_dict_nan = {"var1": data_nan}
    metric.record(data_dict_nan, data_dict_nan, 2)
    metric.record(data_dict_nan, data_dict_nan, 4)
    result_dict = metric.get()
    assert "var1" in result_dict
    result = result_dict["var1"].cpu().numpy()
    # assert tensor is all ones
    assert np.sum(np.isnan(result)) == 2
    assert np.isnan(result[2])
    assert np.isnan(result[4])


def test_single_target_mean_aggregator():
    """
    The area weighted reduced metric should assume that the start
    of a window is always recorded, as we clip it before calling.
    """
    n_sample = 10
    n_time_per_window = 22
    n_window = 3
    nx = 2
    ny = 2
    area_weights = torch.ones(ny).to(fme.get_device())
    torch.manual_seed(0)

    agg = SingleTargetMeanAggregator(
        gridded_operations=LatLonOperations(area_weights),
        n_timesteps=n_time_per_window * n_window,
    )
    data_a = torch.randn(n_sample, n_time_per_window, nx, ny, device=get_device())
    for i in range(n_window):
        data = {"a": data_a[:, i * n_time_per_window : (i + 1) * n_time_per_window]}
        agg.record_batch(data=data, i_time_start=i * n_time_per_window)

    logs = agg.get_logs(label="test")
    assert "test/series" in logs
    ds = agg.get_dataset()
    for i in range(1, data_a.shape[1]):
        raw_variable = data_a[:, i]
        raw_global_mean = raw_variable.mean().cpu().numpy()
        raw_global_std = (
            raw_variable.std(dim=(1, 2), correction=0).mean().cpu().numpy()
        )  # metrics are mean over batch
        np.testing.assert_allclose(
            raw_global_std,
            ds["weighted_std_gen-a"].isel(forecast_step=i).values.item(),
            rtol=1e-5,
        )
        np.testing.assert_allclose(
            raw_global_mean,
            ds["weighted_mean_gen-a"].isel(forecast_step=i).values.item(),
            rtol=1e-5,
        )



from fme.core.config import update_dict_with_dotlist


def test_update_dict_with_dotlist():
    base = {"a": 1, "b": {"c": 2, "d": 3}}
    dotlist = ["a=4", "b.c=0", "b.f.g=5"]
    expected = {"a": 4, "b": {"c": 0, "d": 3, "f": {"g": 5}}}
    merged = update_dict_with_dotlist(base, dotlist)
    assert merged == expected
    assert isinstance(merged, dict)


def test_update_dict_with_dotlist_bool():
    base = {"a": 1}
    dotlist = ["b=true"]
    expected = {"a": 1, "b": True}
    assert update_dict_with_dotlist(base, dotlist) == expected


def test_update_dict_with_none_dotlist():
    base = {"a": 1, "b": {"c": 2, "d": 3}}
    dotlist = None
    assert update_dict_with_dotlist(base, dotlist) == base


def test_update_dict_with_empty_dotlist():
    base = {"a": 1, "b": {"c": 2, "d": 3}}
    dotlist: list[str] = []
    assert update_dict_with_dotlist(base, dotlist) == base



import dataclasses
import datetime
import logging

import numpy as np
import pytest
import torch as th

from fme.ace.models.healpix.healpix_activations import (
    CappedGELUConfig,
    DownsamplingBlockConfig,
)
from fme.ace.models.healpix.healpix_blocks import ConvBlockConfig, RecurrentBlockConfig
from fme.ace.models.healpix.healpix_decoder import UNetDecoder
from fme.ace.models.healpix.healpix_encoder import UNetEncoder
from fme.ace.models.healpix.healpix_layers import HEALPixPadding
from fme.ace.models.healpix.healpix_recunet import HEALPixRecUNet
from fme.ace.registry.hpx import UNetDecoderConfig, UNetEncoderConfig
from fme.ace.stepper import SingleModuleStepperConfig
from fme.core.coordinates import HEALPixCoordinates, HybridSigmaPressureCoordinate
from fme.core.dataset_info import DatasetInfo
from fme.core.device import get_device
from fme.core.normalizer import NormalizationConfig

TIMESTEP = datetime.timedelta(hours=6)
logger = logging.getLogger("__name__")


def fix_random_seeds(seed=0):
    """Fix random seeds for reproducibility"""
    np.random.seed(seed)
    th.manual_seed(seed)
    th.cuda.manual_seed(seed)


def conv_next_block_config(in_channels=3, out_channels=1):
    activation_block_config = CappedGELUConfig(cap_value=10)
    conv_next_block_config = ConvBlockConfig(
        in_channels=in_channels,
        out_channels=out_channels,
        activation=activation_block_config,
        kernel_size=3,
        dilation=1,
        upscale_factor=4,
        block_type="ConvNeXtBlock",
    )
    return conv_next_block_config


def down_sampling_block_config():
    return DownsamplingBlockConfig(pooling=2, block_type="AvgPool")


def encoder_config(
    conv_next_block_config, down_sampling_block_config, n_channels=[136, 68, 34]
):
    return UNetEncoderConfig(
        conv_block=conv_next_block_config,
        down_sampling_block=down_sampling_block_config,
        n_channels=n_channels,
        dilations=[1, 2, 4],
    )


def up_sampling_block_config(in_channels=3, out_channels=1):
    activation_block_config = CappedGELUConfig(cap_value=10)
    transposed_conv_upsample_block_config = ConvBlockConfig(
        in_channels=in_channels,
        out_channels=out_channels,
        activation=activation_block_config,
        stride=2,
        block_type="TransposedConvUpsample",
    )
    return transposed_conv_upsample_block_config


def output_layer_config(in_channels=3, out_channels=2):
    conv_block_config = ConvBlockConfig(
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=1,
        dilation=1,
        n_layers=1,
        block_type="BasicConvBlock",
    )
    return conv_block_config


def recurrent_block_config(in_channels=3):
    recurrent_block_config = RecurrentBlockConfig(
        in_channels=in_channels,
        kernel_size=1,
        block_type="ConvGRUBlock",
    )
    return recurrent_block_config


def decoder_config(
    conv_next_block_config,
    up_sampling_block_config,
    output_layer_config,
    recurrent_block_config,
    n_channels=[34, 68, 136],
):
    decoder_config = UNetDecoderConfig(
        conv_block=conv_next_block_config,
        up_sampling_block=up_sampling_block_config,
        recurrent_block=recurrent_block_config,
        output_layer=output_layer_config,
        n_channels=n_channels,
        dilations=[4, 2, 1],
    )
    return decoder_config


def _test_data():
    # create dummy data
    def generate_test_data(batch_size=8, time_dim=1, channels=7, img_size=16):
        device = get_device()
        test_data = th.randn(batch_size, 12, time_dim * channels, img_size, img_size)
        return test_data.to(device)

    return generate_test_data


def constant_data():
    # create dummy data
    def generate_constant_data(channels=2, img_size=16):
        device = get_device()
        constants = th.randn(12, channels, img_size, img_size)

        return constants.to(device)

    return generate_constant_data


def insolation_data():
    # create dummy data
    def generate_insolation_data(batch_size=8, time_dim=1, img_size=16):
        device = get_device()
        insolation = th.randn(batch_size, 12, time_dim, img_size, img_size)

        return insolation.to(device)

    return generate_insolation_data


@pytest.mark.parametrize("shape", [pytest.param((8, 16))])
def test_hpx_init(shape):
    in_channels = 7
    out_channels = 7
    prognostic_variables = min(in_channels, out_channels)
    n_constants = 1
    decoder_input_channels = 1
    input_time_size = 2
    output_time_size = 4
    device = get_device()

    conv_next_block = conv_next_block_config()
    down_sampling_block = down_sampling_block_config()
    recurrent_block = recurrent_block_config()
    encoder = encoder_config(conv_next_block, down_sampling_block)
    up_sampling_block = up_sampling_block_config()
    output_layer = output_layer_config()
    decoder = decoder_config(
        conv_next_block, up_sampling_block, output_layer, recurrent_block
    )

    hpx_config_data = {
        "type": "HEALPixRecUNet",
        "config": {
            "encoder": dataclasses.asdict(encoder),
            "decoder": dataclasses.asdict(decoder),
            "prognostic_variables": prognostic_variables,
            "n_constants": n_constants,
            "decoder_input_channels": decoder_input_channels,
            "input_time_size": input_time_size,
            "output_time_size": output_time_size,
        },
    }

    stepper_config_data = {
        "builder": hpx_config_data,
        "in_names": ["x"],
        "out_names": ["x"],
        "normalization": dataclasses.asdict(
            NormalizationConfig(
                means={"x": float(np.random.randn(1).item())},
                stds={"x": float(np.random.randn(1).item())},
            )
        ),
    }
    horizontal_coordinates = HEALPixCoordinates(
        th.arange(12), th.arange(8), th.arange(8)
    )
    vertical_coordinate = HybridSigmaPressureCoordinate(
        ak=th.arange(7), bk=th.arange(7)
    ).to(device)
    stepper_config = SingleModuleStepperConfig.from_state(stepper_config_data)
    stepper = stepper_config.get_stepper(
        dataset_info=DatasetInfo(
            horizontal_coordinates=horizontal_coordinates,
            vertical_coordinate=vertical_coordinate,
            timestep=TIMESTEP,
        ),
    )
    assert len(stepper.modules) == 1
    assert type(stepper.modules[0].module) is HEALPixRecUNet


@pytest.mark.parametrize(
    "in_channels, out_channels, n_constants, decoder_input_channels, input_time_size, \
    output_time_size, couplings, expected_exception, expected_message",
    [
        (7, 7, 1, 1, 2, 4, None, None, None),  # Valid case
        (
            7,
            7,
            1,
            1,
            2,
            3,
            None,
            ValueError,
            "'output_time_size' must be a multiple of 'input_time_size'",
        ),  # Bad input and output time dims
        (
            7,
            7,
            0,
            2,
            2,
            3,
            ["t2m", "v10m"],
            NotImplementedError,
            "support for coupled models with no constant field",
        ),  # Couplings with no constants
        (
            7,
            7,
            2,
            0,
            2,
            3,
            ["t2m", "v10m"],
            NotImplementedError,
            "support for coupled models with no decoder",
        ),  # Couplings with no decoder input channels
        (
            7,
            7,
            0,
            0,
            2,
            3,
            None,
            ValueError,
            "'output_time_size' must be a multiple of 'input_time_size'",
        ),  # No constant fields and no decoder
    ],
)
def test_HEALPixRecUNet_initialize(
    in_channels,
    out_channels,
    n_constants,
    decoder_input_channels,
    input_time_size,
    output_time_size,
    couplings,
    expected_exception,
    expected_message,
):
    prognostic_variables = min(out_channels, in_channels)
    conv_next_block = conv_next_block_config()
    up_sampling_block = up_sampling_block_config()
    output_layer = output_layer_config()
    recurrent_block = recurrent_block_config()
    encoder = encoder_config(conv_next_block, down_sampling_block_config())
    decoder = decoder_config(
        conv_next_block, up_sampling_block, output_layer, recurrent_block
    )
    device = get_device()

    if expected_exception:
        with pytest.raises(expected_exception, match=expected_message):
            model = HEALPixRecUNet(
                encoder=encoder,
                decoder=decoder,
                input_channels=in_channels,
                output_channels=out_channels,
                prognostic_variables=prognostic_variables,
                n_constants=n_constants,
                decoder_input_channels=decoder_input_channels,
                input_time_size=input_time_size,
                output_time_size=output_time_size,
                couplings=couplings,
            ).to(device)
    else:
        model = HEALPixRecUNet(
            encoder=encoder,
            decoder=decoder,
            input_channels=in_channels,
            output_channels=out_channels,
            prognostic_variables=prognostic_variables,
            n_constants=n_constants,
            decoder_input_channels=decoder_input_channels,
            input_time_size=input_time_size,
            output_time_size=output_time_size,
            couplings=couplings,
        ).to(device)
        assert isinstance(model, HEALPixRecUNet)


def test_HEALPixRecUNet_integration_steps():
    in_channels = 2
    out_channels = 2
    prognostic_variables = min(out_channels, in_channels)
    n_constants = 1
    decoder_input_channels = 0
    input_time_size = 2
    output_time_size = 4
    device = get_device()

    conv_next_block = conv_next_block_config()
    up_sampling_block = up_sampling_block_config()
    output_layer = output_layer_config()
    recurrent_block = recurrent_block_config()
    encoder = encoder_config(conv_next_block, down_sampling_block_config())
    decoder = decoder_config(
        conv_next_block, up_sampling_block, output_layer, recurrent_block
    )

    model = HEALPixRecUNet(
        encoder=encoder,
        decoder=decoder,
        input_channels=in_channels,
        output_channels=out_channels,
        prognostic_variables=prognostic_variables,
        n_constants=n_constants,
        decoder_input_channels=decoder_input_channels,
        input_time_size=input_time_size,
        output_time_size=output_time_size,
    ).to(device)

    assert model.integration_steps == output_time_size // input_time_size


def test_HEALPixRecUNet_reset(very_fast_only: bool):
    if very_fast_only:
        pytest.skip("Skipping non-fast tests")
    # create a smaller version of the dlwp healpix model
    in_channels = 3
    out_channels = 3
    prognostic_variables = min(out_channels, in_channels)
    n_constants = 2
    decoder_input_channels = 1
    input_time_size = 2
    output_time_size = 4
    size = 16
    device = get_device()

    conv_next_block = conv_next_block_config()
    up_sampling_block = up_sampling_block_config()
    output_layer = output_layer_config()
    recurrent_block = recurrent_block_config()
    encoder = encoder_config(conv_next_block, down_sampling_block_config())
    decoder = decoder_config(
        conv_next_block, up_sampling_block, output_layer, recurrent_block
    )

    fix_random_seeds(seed=42)
    x = _test_data()(time_dim=input_time_size, channels=in_channels, img_size=size)
    decoder_inputs = insolation_data()(time_dim=input_time_size, img_size=size)
    constants = constant_data()(channels=n_constants, img_size=size)
    batch_size = x.shape[0]
    constants = constants.unsqueeze(0).repeat(batch_size, 1, 1, 1, 1)
    inputs = th.concat(
        (x, decoder_inputs, constants), dim=-3
    )  # [x, decoder_inputs, constants]

    model = HEALPixRecUNet(
        encoder=encoder,
        decoder=decoder,
        input_channels=in_channels,
        output_channels=out_channels,
        prognostic_variables=prognostic_variables,
        n_constants=n_constants,
        decoder_input_channels=decoder_input_channels,
        input_time_size=input_time_size,
        output_time_size=output_time_size,
        enable_healpixpad=False,
        delta_time="6h",
    ).to(device)

    out_var = model(inputs)
    model.reset()

    assert compare_output(out_var, model(inputs))


# Checks the model can perform a forward class on various input configurations
# [full inputs, no decoder inputs, no constant inputs]
@pytest.mark.parametrize(
    "inputs_config, in_channels, decoder_input_channels, \
        out_channels, input_time_size, output_time_size, n_constants, size",
    [
        ([0, 1, 2], 3, 1, 3, 2, 4, 2, 16),  # full inputs
        ([0, 2], 3, 0, 3, 2, 4, 2, 16),  # no decoder inputs
        ([0, 1], 3, 1, 3, 2, 4, 0, 16),  # no constant inputs
    ],
)
def test_HEALPixRecUNet_forward(
    inputs_config,
    in_channels,
    decoder_input_channels,
    out_channels,
    input_time_size,
    output_time_size,
    n_constants,
    size,
    very_fast_only: bool,
):
    if very_fast_only:
        pytest.skip("Skipping non-fast tests")
    prognostic_variables = min(out_channels, in_channels)
    device = get_device()
    conv_next_block = conv_next_block_config()
    up_sampling_block = up_sampling_block_config()
    output_layer = output_layer_config()
    recurrent_block = recurrent_block_config()
    encoder = encoder_config(conv_next_block, down_sampling_block_config())
    decoder = decoder_config(
        conv_next_block, up_sampling_block, output_layer, recurrent_block
    )

    fix_random_seeds(seed=42)
    x = _test_data()(time_dim=input_time_size, channels=in_channels, img_size=size)
    batch_size = x.shape[0]

    if decoder_input_channels > 0:
        decoder_inputs = insolation_data()(time_dim=input_time_size, img_size=size)
    else:
        decoder_inputs = insolation_data()(time_dim=0, img_size=size)
    constants = constant_data()(channels=n_constants, img_size=size)
    constants = constants.unsqueeze(0).repeat(batch_size, 1, 1, 1, 1)

    all_inputs = [x, decoder_inputs, constants]
    inputs = th.concat(all_inputs, dim=-3)

    model = HEALPixRecUNet(
        encoder=encoder,
        decoder=decoder,
        input_channels=in_channels,
        output_channels=out_channels,
        prognostic_variables=prognostic_variables,
        n_constants=n_constants,
        decoder_input_channels=decoder_input_channels,
        input_time_size=input_time_size,
        output_time_size=output_time_size,
        enable_healpixpad=False,
        delta_time="6h",
    ).to(device)
    model(inputs)


# pragma mark - encoder


def test_UNetEncoder_initialize():
    device = get_device()
    channels = 2
    n_channels = (16, 32, 64)

    # Dicts for block configs used by encoder
    conv_block_config = ConvBlockConfig(
        in_channels=channels,
        block_type="ConvNeXtBlock",
    )
    down_sampling_block_config = DownsamplingBlockConfig(
        pooling=2, block_type="MaxPool"
    )

    encoder = UNetEncoder(
        conv_block=conv_block_config,
        down_sampling_block=down_sampling_block_config,
        n_channels=n_channels,
        input_channels=channels,
    ).to(device)
    assert isinstance(encoder, UNetEncoder)

    # with dilations
    encoder = UNetEncoder(
        conv_block=conv_block_config,
        down_sampling_block=down_sampling_block_config,
        n_channels=n_channels,
        input_channels=channels,
        dilations=[1, 1, 1],
    ).to(device)
    assert isinstance(encoder, UNetEncoder)


def test_UNetEncoder_forward():
    channels = 2
    hw_size = 16
    b_size = 12
    n_channels = (16, 32, 64)
    device = get_device()

    # block configs used by encoder
    conv_block_config = ConvBlockConfig(
        in_channels=channels,
        block_type="ConvNeXtBlock",
    )
    down_sampling_block_config = DownsamplingBlockConfig(
        pooling=2, block_type="MaxPool"
    )
    encoder = UNetEncoder(
        conv_block=conv_block_config,
        down_sampling_block=down_sampling_block_config,
        n_channels=n_channels,
        input_channels=channels,
    ).to(device)

    tensor_size = [b_size, channels, hw_size, hw_size]
    invar = th.rand(tensor_size).to(device)
    outvar = encoder(invar)

    # doesn't do anything
    encoder.reset()

    # outvar is a module list
    for idx, out_tensor in enumerate(outvar):
        # verify the channels and h dim are correct
        assert out_tensor.shape[1] == n_channels[idx]
        # default behaviour is to half the h/w size after first
        assert out_tensor.shape[2] == tensor_size[2] // (2**idx)


def test_UNetEncoder_reset():
    channels = 2
    n_channels = (16, 32, 64)
    device = get_device()

    # Dicts for block configs used by encoder
    conv_block_config = ConvBlockConfig(
        in_channels=channels,
        block_type="ConvNeXtBlock",
    )
    down_sampling_block_config = DownsamplingBlockConfig(
        pooling=2,
        block_type="MaxPool",
    )
    encoder = UNetEncoder(
        conv_block=conv_block_config,
        down_sampling_block=down_sampling_block_config,
        n_channels=n_channels,
        input_channels=channels,
    ).to(device)

    # doesn't do anything
    encoder.reset()
    assert isinstance(encoder, UNetEncoder)


def test_UNetDecoder_initilization():
    in_channels = 2
    out_channels = 1
    n_channels = (64, 32, 16)
    device = get_device()

    # Dicts for block configs used by decoder
    conv_block_config = ConvBlockConfig(
        in_channels=in_channels, out_channels=out_channels, block_type="ConvNeXtBlock"
    )
    up_sampling_block_config = ConvBlockConfig(
        in_channels=in_channels,
        out_channels=out_channels,
        stride=2,
        block_type="TransposedConvUpsample",
    )

    output_layer_config = ConvBlockConfig(
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=1,
        dilation=1,
        n_layers=1,
        block_type="ConvNeXtBlock",
    )

    recurrent_block_config = RecurrentBlockConfig(
        in_channels=2,
        kernel_size=1,
        block_type="ConvGRUBlock",
    )

    decoder = UNetDecoder(
        conv_block=conv_block_config,
        up_sampling_block=up_sampling_block_config,
        output_layer=output_layer_config,
        recurrent_block=recurrent_block_config,
        n_channels=n_channels,
    ).to(device)

    assert isinstance(decoder, UNetDecoder)

    # without the recurrent block and with dilations
    decoder = UNetDecoder(
        conv_block=conv_block_config,
        up_sampling_block=up_sampling_block_config,
        output_layer=output_layer_config,
        recurrent_block=None,
        n_channels=n_channels,
        dilations=[1, 1, 1],
    ).to(device)
    assert isinstance(decoder, UNetDecoder)


def test_UNetDecoder_forward():
    in_channels = 2
    out_channels = 1
    hw_size = 32
    b_size = 12
    n_channels = (64, 32, 16)
    device = get_device()

    # Dicts for block configs used by decoder
    conv_block_config = ConvBlockConfig(
        in_channels=in_channels, out_channels=out_channels, block_type="ConvNeXtBlock"
    )
    up_sampling_block_config = ConvBlockConfig(
        in_channels=in_channels,
        out_channels=out_channels,
        stride=2,
        block_type="TransposedConvUpsample",
    )
    output_layer_config = ConvBlockConfig(
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=1,
        dilation=1,
        n_layers=1,
        block_type="BasicConvBlock",
    )
    recurrent_block_config = RecurrentBlockConfig(
        in_channels=2,
        kernel_size=1,
        block_type="ConvGRUBlock",
    )

    decoder = UNetDecoder(
        conv_block=conv_block_config,
        up_sampling_block=up_sampling_block_config,
        output_layer=output_layer_config,
        recurrent_block=recurrent_block_config,
        n_channels=n_channels,
    ).to(device)

    output_2_size = th.Size([b_size, out_channels, hw_size, hw_size])

    # build the list of tensors for the decoder
    invars = []
    # decoder has an algorithm that goes back to front
    for idx in range(len(n_channels) - 1, -1, -1):
        tensor_size = [b_size, n_channels[idx], hw_size, hw_size]
        invars.append(th.rand(tensor_size).to(device))
        hw_size = hw_size // 2

    outvar = decoder(invars)
    assert outvar.shape == output_2_size

    # make sure history is taken into account with ConvGRU
    outvar_hist = decoder(invars)
    assert not compare_output(outvar, outvar_hist)

    # check with no recurrent
    decoder = UNetDecoder(
        conv_block=conv_block_config,
        up_sampling_block=up_sampling_block_config,
        output_layer=output_layer_config,
        recurrent_block=None,
        n_channels=n_channels,
        dilations=[1, 1, 1],
    ).to(device)

    outvar = decoder(invars)
    assert outvar.shape == output_2_size


def test_UNetDecoder_reset():
    in_channels = 2
    out_channels = 1
    hw_size = 32
    b_size = 12
    n_channels = (64, 32, 16)
    device = get_device()

    # Dicts for block configs used by decoder
    conv_block = ConvBlockConfig(in_channels=in_channels, block_type="ConvNeXtBlock")
    up_sampling_block = ConvBlockConfig(
        in_channels=in_channels,
        out_channels=out_channels,
        block_type="TransposedConvUpsample",
    )
    output_layer = ConvBlockConfig(
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=1,
        dilation=1,
        n_layers=1,
        block_type="BasicConvBlock",
    )

    recurrent_block = RecurrentBlockConfig(
        in_channels=2, kernel_size=1, block_type="ConvLSTMBlock"
    )

    decoder = UNetDecoder(
        conv_block=conv_block,
        up_sampling_block=up_sampling_block,
        output_layer=output_layer,
        recurrent_block=recurrent_block,
        n_channels=n_channels,
    ).to(device)

    # build the list of tensors for the decoder
    invars = []
    # decoder has an algorithm that goes back to front
    for idx in range(len(n_channels) - 1, -1, -1):
        tensor_size = [b_size, n_channels[idx], hw_size, hw_size]
        invars.append(th.rand(tensor_size).to(device))
        hw_size = hw_size // 2

    outvar = decoder(invars)

    # make sure history is taken into account with ConvGRU
    outvar_hist = decoder(invars)
    assert not compare_output(outvar, outvar_hist)

    # make sure after reset we get the same result
    decoder.reset()
    outvar_reset = decoder(invars)
    assert compare_output(outvar, outvar_reset)

    # test reset without recurrent block
    decoder = UNetDecoder(
        conv_block=conv_block,
        up_sampling_block=up_sampling_block,
        output_layer=output_layer,
        recurrent_block=None,
        n_channels=n_channels,
    ).to(device)

    outvar = decoder(invars)

    # without the recurrent block should be the same
    outvar_hist = decoder(invars)
    assert compare_output(outvar, outvar_hist)

    # make sure after reset we get the same result
    decoder.reset()
    outvar_reset = decoder(invars)
    assert compare_output(outvar, outvar_reset)


def compare_output(
    output_1: th.Tensor | tuple[th.Tensor, ...],
    output_2: th.Tensor | tuple[th.Tensor, ...],
    rtol: float = 1e-5,
    atol: float = 1e-5,
) -> bool:
    """Compares model outputs and returns if they are the same

    Args
        output_1: First item to compare
        output_2: Second item to compare
        rtol: Relative tolerance of error allowed, by default 1e-5
        atol: Absolute tolerance of error allowed, by default 1e-5

    Returns:
        If outputs are the same
    """
    # Output of tensor
    if isinstance(output_1, th.Tensor):
        return th.allclose(output_1, output_2, rtol, atol)
    # Output of tuple of tensors
    elif isinstance(output_1, tuple):
        # Loop through tuple of outputs
        for i, (out_1, out_2) in enumerate(zip(output_1, output_2)):
            # If tensor use allclose
            if isinstance(out_1, th.Tensor):
                if not th.allclose(out_1, out_2, rtol, atol):
                    logger.warning(f"Failed comparison between outputs {i}")
                    logger.warning(f"Max Difference: {th.amax(th.abs(out_1 - out_2))}")
                    logger.warning(f"Difference: {out_1 - out_2}")
                    return False
            # Otherwise assume primative
            else:
                if not out_1 == out_2:
                    return False
    elif isinstance(output_1, list | tuple) and isinstance(output_2, list | tuple):
        if len(output_1) != len(output_2):
            print(
                f"Length mismatch: output_1 {len(output_1)}, output_2 {len(output_2)}"
            )
            return False
        for a, e in zip(output_1, output_2):
            if not compare_output(a, e):
                return False
        return True
    elif isinstance(output_1, dict) and isinstance(output_2, dict):
        if output_1.keys() != output_2.keys():
            print(
                f"Keys mismatch: output_1 keys {output_1.keys()}, ",
                f"output_2 keys {output_2.keys()}",
            )
            return False
        for key in output_1:
            if not compare_output(output_1[key], output_2[key]):
                return False
        return True
    # Unsupported output type
    else:
        logger.error(
            "Model returned invalid type for unit test, \
            should be th.Tensor or Tuple[th.Tensor]"
        )
        return False
    return True


@pytest.fixture
def mock_data():
    """
    Create a mock tensor with known values for 12 HEALPix faces.
    Shape: [B=1, F=12, C=1, H=4, W=4] - a single batch with 12 faces, one channel,
        4x4 grid
    """
    return th.arange(1, 12 * 4 * 4 + 1, dtype=th.float32).reshape((12, 1, 4, 4))


@pytest.fixture
def healpix_padding():
    padding = 2

    """Instantiate HEALPixPadding with the specified padding."""
    return HEALPixPadding(padding=padding, enable_nhwc=False)


def test_healpix_padding_pn(healpix_padding):
    """
    Test north hemisphere padding.
    This test checks if each northern face is padded correctly with its neighbors.
    """
    padding = 2

    # Mock the neighbor faces, assuming they are 4x4 for simplicity.
    face = th.ones((1, 1, 4, 4))
    top = face
    top_left = face * 2
    left = face * 3
    bottom_left = face * 4
    bottom = face * 5
    bottom_right = face * 6
    right = face * 7
    top_right = face * 8

    # Run the pn function directly for controlled testing
    padded = healpix_padding.pn(
        face, top, top_left, left, bottom_left, bottom, bottom_right, right, top_right
    )

    # Check if padding applied matches expected size: 4 + 2*padding in each dimension
    assert padded.shape[-2:] == (4 + 2 * padding, 4 + 2 * padding)
    # North padding (top two rows) should match `top`
    assert th.all(padded[..., :padding, padding:-padding] == top[..., -padding:, :])
    # Northwest corner padding should match `top_left`
    assert th.all(
        padded[..., :padding, :padding] == top_left[..., -padding:, -padding:]
    )
    # Northeast corner padding should match `top_right`
    assert th.all(
        padded[..., :padding, -padding:] == top_right[..., -padding:, :padding]
    )


def test_healpix_padding_pe(healpix_padding):
    """
    Test equatorial face padding to ensure proper padding alignment with
    equatorial neighbors.
    """
    padding = 2

    # Mock the neighbor faces, assuming they are 4x4 for simplicity.
    face = th.ones((1, 1, 4, 4))
    top = face
    top_left = face * 2
    left = face * 3
    bottom_left = face * 4
    bottom = face * 5
    bottom_right = face * 6
    right = face * 7
    top_right = face * 8

    # Run the pe function directly for controlled testing
    padded = healpix_padding.pe(
        face, top, top_left, left, bottom_left, bottom, bottom_right, right, top_right
    )

    # Check if padding applied matches expected size: 4 + 2*padding in each dimension
    assert padded.shape[-2:] == (4 + 2 * padding, 4 + 2 * padding)
    # Left padding (left two columns) should match `left`
    assert th.all(padded[..., padding:-padding, :padding] == left[..., :, -padding:])
    # Right padding (right two columns) should match `right`
    assert th.all(padded[..., padding:-padding, -padding:] == right[..., :, :padding])


def test_healpix_padding_ps(healpix_padding):
    """
    Test south hemisphere padding.
    This test checks if each southern face is padded correctly with its neighbors.
    """
    padding = 2

    # Mock the neighbor faces, assuming they are 4x4 for simplicity.
    face = th.ones((1, 1, 4, 4))
    top = face
    top_left = face * 2
    left = face * 3
    bottom_left = face * 4
    bottom = face * 5
    bottom_right = face * 6
    right = face * 7
    top_right = face * 8

    # Run the ps function directly for controlled testing
    padded = healpix_padding.ps(
        face, top, top_left, left, bottom_left, bottom, bottom_right, right, top_right
    )

    # Check if padding applied matches expected size: 4 + 2*padding in each dimension
    assert padded.shape[-2:] == (4 + 2 * padding, 4 + 2 * padding)
    # South padding (bottom two rows) should match `bottom`
    assert th.all(padded[..., -padding:, padding:-padding] == bottom[..., :padding, :])
    # Southwest corner padding should match `bottom_left`
    assert th.all(
        padded[..., -padding:, :padding] == bottom_left[..., :padding, -padding:]
    )
    # Southeast corner padding should match `bottom_right`
    assert th.all(
        padded[..., -padding:, -padding:] == bottom_right[..., :padding, :padding]
    )


def test_healpix_padding_forward(healpix_padding, mock_data):
    """
    Full integration test for HEALPixPadding's forward method.
    Checks that all faces receive the correct padding.
    """
    padding = 2

    padded_data = healpix_padding(mock_data)

    expected_height = mock_data.shape[-2] + 2 * padding
    expected_width = mock_data.shape[-1] + 2 * padding
    assert padded_data.shape[-2:] == (expected_height, expected_width)

    # TODO: check each face to confirm padding values in each region
    for face_idx in range(mock_data.shape[0]):
        face_data = padded_data[face_idx, 0]

        # the region where original data should reside
        data_region = face_data[padding : padding + 4, padding : padding + 4]
        original_data = mock_data[face_idx, 0]  # Original face data
        assert (
            data_region == original_data
        ).all(), f"Data region on face {face_idx} has been altered."



import torch

from fme.core.packer import Packer


def test_pack_singleton_channels():
    data = {
        "u10": torch.randn(2, 4, 8),
        "v10": torch.randn(2, 4, 8),
        "w10": torch.randn(2, 4, 8),
    }
    p = Packer(names=["u10", "v10"])
    packed = p.pack(data, axis=1)
    assert packed.shape == (2, 2, 4, 8)
    assert torch.allclose(packed[:, 0, :, :], data["u10"])
    assert torch.allclose(packed[:, 1, :, :], data["v10"])


def test_unpack():
    tensor = torch.randn(2, 2, 4, 8)
    p = Packer(names=["u10", "v10"])
    unpacked = p.unpack(tensor, axis=1)
    assert len(unpacked) == 2
    assert torch.allclose(unpacked["u10"], tensor[:, 0, :, :])
    assert torch.allclose(unpacked["v10"], tensor[:, 1, :, :])


def test_unpack_first_axis():
    tensor = torch.randn(2, 2, 4, 8)
    p = Packer(names=["u10", "v10"])
    unpacked = p.unpack(tensor, axis=0)
    assert len(unpacked) == 2
    assert torch.allclose(unpacked["u10"], tensor[0, :, :, :])
    assert torch.allclose(unpacked["v10"], tensor[1, :, :, :])



import torch

from fme.ace.aggregator.one_step.ensemble import CRPSMetric, SSRBiasMetric


def get_tensor(shape):
    return torch.randn(*shape)


def test_crps_metric_gives_correct_shape():
    metric = CRPSMetric()
    n_batch = 10
    n_sample = 2
    n_time = 3
    n_y = 4
    n_x = 5
    target = get_tensor((n_batch, 1, n_time, n_y, n_x))
    gen = get_tensor((n_batch, n_sample, n_time, n_y, n_x))
    metric.record(target=target, gen=gen)
    got = metric.get()
    assert isinstance(got, torch.Tensor)
    assert got.shape == (n_y, n_x)


def test_ssr_metric_gives_correct_shape():
    torch.manual_seed(0)
    metric = SSRBiasMetric()
    n_batch = 10
    n_sample = 2
    n_time = 3
    n_y = 4
    n_x = 5
    target = get_tensor((n_batch, 1, n_time, n_y, n_x))
    gen = get_tensor((n_batch, n_sample, n_time, n_y, n_x))
    metric.record(target=target, gen=gen)
    got = metric.get()
    assert isinstance(got, torch.Tensor)
    assert got.shape == (n_y, n_x)


def test_ssr_bias_metric_unbiased():
    torch.manual_seed(0)
    metric = SSRBiasMetric()
    n_batch = 5000
    n_sample = 2
    n_time = 3
    n_y = 4
    n_x = 5
    # identical distribution for target and gen
    target = get_tensor((n_batch, 1, n_time, n_y, n_x))
    gen = get_tensor((n_batch, n_sample, n_time, n_y, n_x))
    metric.record(target=target, gen=gen)
    got = metric.get()
    assert isinstance(got, torch.Tensor)
    assert got.shape == (n_y, n_x)
    torch.testing.assert_close(got.mean(), torch.tensor(0.0), atol=1e-2, rtol=0.0)


def test_ssr_bias_metric_doubled_spread():
    torch.manual_seed(0)
    metric = SSRBiasMetric()
    n_batch = 5000
    n_sample = 2
    n_time = 3
    n_y = 4
    n_x = 5
    target = get_tensor((n_batch, 1, n_time, n_y, n_x))
    gen = get_tensor((n_batch, n_sample, n_time, n_y, n_x)) * 2
    metric.record(target=target, gen=gen)
    got = metric.get()
    assert isinstance(got, torch.Tensor)
    assert got.shape == (n_y, n_x)
    torch.testing.assert_close(got.mean(), torch.tensor(1.0), atol=1e-2, rtol=0.0)



import torch

from fme.ace.aggregator.one_step.deterministic import OneStepDeterministicAggregator
from fme.core.coordinates import LatLonCoordinates
from fme.core.dataset_info import DatasetInfo


def test__get_loss_scaled_mse_components():
    loss_scaling = {
        "a": torch.tensor(1.0),
        "b": torch.tensor(0.5),
    }
    nx, ny = 10, 10
    lat_lon_coordinates = LatLonCoordinates(torch.arange(nx), torch.arange(ny))
    # keep area weights ones for simplicity
    lat_lon_coordinates._area_weights = torch.ones(nx, ny)
    ds_info = DatasetInfo(horizontal_coordinates=lat_lon_coordinates)
    agg = OneStepDeterministicAggregator(
        ds_info,
        loss_scaling=loss_scaling,
        save_diagnostics=False,
    )

    logs = {
        "test/mean/weighted_rmse/a": 1.0,
        "test/mean/weighted_rmse/b": 4.0,
        "test/mean/weighted_rmse/c": 0.0,
    }
    result = agg._get_loss_scaled_mse_components(logs, "test")
    scaled_squared_errors_sum = (1.0 / 1.0) ** 2 + (4.0 / 0.5) ** 2
    assert (
        result["test/mean/mse_fractional_components/a"] == 1 / scaled_squared_errors_sum
    )
    assert (
        result["test/mean/mse_fractional_components/b"]
        == 64 / scaled_squared_errors_sum
    )
    assert "test/mean/mse_fractional_components/c" not in result



import dataclasses
from collections.abc import Callable, Mapping

# we use Type to distinguish from type attr of CorrectorSelector
from typing import Any, ClassVar, Type, TypeVar  # noqa: UP035

from .registry import Registry

# Note we can either type hint register, which prevents registering
# from overwriting the type with a generic type or Any, _or_ we can
# type hint the registry attribute below, which prevents `instance`
# in `build` from being typed as `Any`. We cannot do both, unfortunately.
T = TypeVar("T")


@dataclasses.dataclass
class CorrectorSelector:
    """
    A dataclass containing all the information needed to build a
    CorrectorConfigProtocol, including the type of the CorrectorConfigProtocol
    and the data needed to build it.

    This is helpful as CorrectorSelector can be serialized and deserialized
    without any additional information, whereas to load a
    CorrectorConfigProtocol you would need to know the type of the
    CorrectorConfigProtocol being loaded.

    It is also convenient because CorrectorSelector is a single class that can
    be used to represent any CorrectorConfigProtocol, whereas
    CorrectorConfigProtocol is a protocol that can be implemented by many
    different classes.

    Parameters:
        type: the type of the CorrectorConfigProtocol
        config: data for a CorrectorConfigProtocol instance of the indicated type
    """

    type: str
    config: Mapping[str, Any]
    # Note this registry is no longer used, but is kept for backwards compatibility.
    # In practice the corrector is selected based on the type of the vertical
    # coordinate in the loaded data.
    registry: ClassVar[Registry] = Registry()

    def __post_init__(self):
        if not isinstance(self.registry, Registry):
            raise ValueError("CorrectorSelector.registry should not be set manually")

    @classmethod
    def register(cls, type_name) -> Callable[[Type[T]], Type[T]]:  # noqa: UP006
        return cls.registry.register(type_name)

    @classmethod
    def get_available_types(cls):
        """This class method is used to expose all available types of Correctors."""
        return cls(type="", config={}).registry._types.keys()
